# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import Enum, IntFlag
from typing import Any, List, Optional, Tuple, Union

import numpy as np
from pyglm.glm import inverse, mat3, mat4x3, rotate, transpose, vec3, vec4
from pyxpg import (
    AllocType,
    Attachment,
    BlendFactor,
    BlendOp,
    Buffer,
    BufferUsageFlags,
    ComputePipeline,
    CullMode,
    Depth,
    DescriptorSet,
    DescriptorSetBinding,
    DescriptorType,
    DeviceFeatures,
    Format,
    FrontFace,
    GraphicsPipeline,
    ImageLayout,
    ImageUsageFlags,
    InputAssembly,
    MemoryUsage,
    PipelineStage,
    PipelineStageFlags,
    PolygonMode,
    PrimitiveTopology,
    PushConstantsRange,
    Rasterization,
    Shader,
    Stage,
    VertexAttribute,
    VertexBinding,
    VertexInputRate,
)

from .geometry import concatenate_meshes, create_arrow, create_sphere, transform_mesh
from .gpu_sorting import GpuSortingPipeline, SortDataType, SortOptions
from .marching_cubes import BLOCK_SIZE_X, BLOCK_SIZE_Y, BLOCK_SIZE_Z, MarchingCubesPipeline
from .materials import ColorMaterial, DiffuseMaterial, Material
from .property import BufferProperty, ImageProperty, as_image_property
from .renderer import Renderer
from .renderer_frame import RendererFrame
from .scene import Object, Object3D
from .utils.descriptors import create_descriptor_layout_pool_and_set, create_descriptor_layout_pool_and_sets_ringbuffer
from .utils.gpu import (
    AccelerationStructureInstanceInfo,
    align_up,
    cull_mode_opposite_face,
    div_ceil,
    div_round_up,
    view_bytes,
)
from .viewport import Viewport


class ColormapKind(Enum):
    AUTUMN = 0
    BONE = 1
    COOL = 2
    COPPER = 3
    HOT = 4
    HSV = 5
    JET = 6
    SPRING = 7
    SUMMER = 8
    WINTER = 9


@dataclass
class Colormap:
    color: ColormapKind = ColormapKind.JET
    range_min: float = 0.0
    range_max: float = 1.0


@dataclass
class ColormapDistanceToPoint(Colormap):
    point: vec3 = vec3(0, 0, 0)


@dataclass
class ColormapDistanceToPlane(Colormap):
    normal: vec3 = vec3(0, 0, 1)


class Points(Object3D):
    def __init__(
        self,
        points: BufferProperty,
        colors: Optional[BufferProperty] = None,
        uniform_color: Optional[int] = None,
        colormap: Optional[Colormap] = None,
        point_size: Union[BufferProperty, float] = 1.0,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        if uniform_color is None and colors is None and colormap is None:
            uniform_color = 0xFFCCCCCC

        if uniform_color is not None:
            if colormap is not None or colors is not None:
                raise ValueError("Only one of uniform_color, colormap and colors can be not None")
            self.constants_dtype = np.dtype(
                {
                    "transform": (np.dtype((np.float32, (3, 4))), 0),
                    "point_size": (np.float32, 48),
                    "color": (np.uint32, 52),
                }
            )  # type: ignore
        elif colors is not None:
            if colormap is not None:
                raise ValueError("Only one of uniform_color, colormap and colors can be not None")
            self.constants_dtype = np.dtype(
                {
                    "transform": (np.dtype((np.float32, (3, 4))), 0),
                    "point_size": (np.float32, 48),
                }
            )  # type: ignore
        elif colormap is not None:
            self.constants_dtype = np.dtype(
                {
                    "transform": (np.dtype((np.float32, (3, 4))), 0),
                    "point_size": (np.float32, 48),
                    "colormap_measure": (np.uint32, 52),
                    "range_min": (np.float32, 56),
                    "range_inv_delta": (np.float32, 60),
                    "point_or_normal": (np.dtype((np.float32, (3,))), 64),
                }
            )  # type: ignore

        self.constants = np.zeros((1,), self.constants_dtype)

        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)
        self.point_size = self.add_buffer_property(point_size, np.float32, name="point_size")
        self.points = self.add_buffer_property(points, np.float32, (-1, 3), name="points").use_gpu(
            BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
        )
        self.colors = (
            self.add_buffer_property(colors, np.uint32, (-1,), name="colors").use_gpu(
                BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
            )
            if colors is not None
            else None
        )
        self.uniform_color = uniform_color
        self.colormap = colormap

    def create(self, r: Renderer) -> None:
        vertex_bindings = [VertexBinding(0, 12, VertexInputRate.VERTEX)]
        vertex_attributes = [VertexAttribute(0, 0, Format.R32G32B32_SFLOAT)]

        defines = []
        if self.uniform_color is not None:
            defines.append(("UNIFORM_COLOR", ""))
        elif self.colors is not None:
            defines.append(("VERTEX_COLORS", ""))
            vertex_bindings.append(VertexBinding(1, 4, VertexInputRate.VERTEX))
            vertex_attributes.append(VertexAttribute(1, 1, Format.R32_UINT))
        elif self.colormap is not None:
            defines.append(("COLORMAP", ""))

        vert = r.compile_builtin_shader("3d/points.slang", "vertex_main", defines=defines)
        frag = r.compile_builtin_shader("3d/points.slang", "pixel_main", defines=defines)

        # Instantiate the pipeline using the compiled shaders
        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=vertex_bindings,
            vertex_attributes=vertex_attributes,
            input_assembly=InputAssembly(PrimitiveTopology.POINT_LIST),
            samples=r.msaa_samples,
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"] = mat4x3(self.current_transform_matrix)

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        self.constants["point_size"] = self.point_size.get_current()
        if self.uniform_color is not None:
            self.constants["color"] = self.uniform_color
        elif self.colormap is not None:
            if isinstance(self.colormap, ColormapDistanceToPoint):
                measure = 0
                self.constants["point_or_normal"][:3] = self.colormap.point
            elif isinstance(self.colormap, ColormapDistanceToPlane):
                measure = 1
                self.constants["point_or_normal"][:3] = self.colormap.normal
            else:
                raise ValueError("colormap must be of type ColormapDistanceToPlane or ColormapDistanceToPoint")
            self.constants["colormap_measure"] = (measure << 16) | self.colormap.color.value
            self.constants["range_min"] = self.colormap.range_min
            self.constants["range_inv_delta"] = 1.0 / (self.colormap.range_max - self.colormap.range_min)

        points = self.points.get_current_gpu()

        vertex_buffers = [points.buffer_and_offset()]
        if self.colors is not None:
            vertex_buffers.append(self.colors.get_current_gpu().buffer_and_offset())

        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=vertex_buffers,
            descriptor_sets=[
                scene_descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )
        frame.cmd.draw(points.size // 12)


class Lines(Object3D):
    def __init__(
        self,
        lines: BufferProperty,
        colors: BufferProperty,
        line_width: Union[BufferProperty, float] = 1.0,
        is_strip: bool = False,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)

        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)
        self.is_strip = is_strip
        self.lines = self.add_buffer_property(lines, np.float32, (-1, 3), name="lines").use_gpu(
            BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
        )
        self.colors = self.add_buffer_property(colors, np.uint32, (-1,), name="colors").use_gpu(
            BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
        )
        self.line_width = self.add_buffer_property(line_width, np.float32, name="line_width")

    def create(self, r: Renderer) -> None:
        vert = r.compile_builtin_shader("3d/basic.slang", "vertex_main")
        frag = r.compile_builtin_shader("3d/basic.slang", "pixel_main")

        # Instantiate the pipeline using the compiled shaders
        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=[
                VertexBinding(0, 12, VertexInputRate.VERTEX),
                VertexBinding(1, 4, VertexInputRate.VERTEX),
            ],
            vertex_attributes=[
                VertexAttribute(0, 0, Format.R32G32B32_SFLOAT),
                VertexAttribute(1, 1, Format.R32_UINT),
            ],
            rasterization=Rasterization(dynamic_line_width=True),
            input_assembly=InputAssembly(
                PrimitiveTopology.LINE_STRIP if self.is_strip else PrimitiveTopology.LINE_LIST
            ),
            samples=r.msaa_samples,
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"] = mat4x3(self.current_transform_matrix)

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        lines = self.lines.get_current_gpu()
        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=[
                lines.buffer_and_offset(),
                self.colors.get_current_gpu().buffer_and_offset(),
            ],
            descriptor_sets=[
                scene_descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )
        frame.cmd.set_line_width(
            self.line_width.get_current().item() if renderer.device.features & DeviceFeatures.WIDE_LINES else 1.0
        )

        frame.cmd.draw(lines.size // 12)


class Mesh(Object3D):
    def __init__(
        self,
        positions: BufferProperty,
        indices: Optional[BufferProperty] = None,
        normals: Optional[BufferProperty] = None,
        tangents: Optional[BufferProperty] = None,
        uvs: Optional[BufferProperty] = None,
        vertex_colors: Optional[BufferProperty] = None,
        instance_positions: Optional[BufferProperty] = None,
        primitive_topology: PrimitiveTopology = PrimitiveTopology.TRIANGLE_LIST,
        cull_mode: CullMode = CullMode.NONE,
        front_face: FrontFace = FrontFace.COUNTER_CLOCKWISE,
        polygon_mode: PolygonMode = PolygonMode.FILL,
        material: Optional[Material] = None,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.primitive_topology = primitive_topology
        self.cull_mode = cull_mode
        self.front_face = front_face
        self.polygon_mode = polygon_mode
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
                "normal_matrix": (np.dtype((np.float32, (3, 4))), 48),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)

        if material is None:
            default_color = (0.5, 0.5, 0.5) if vertex_colors is None else (1.0, 1.0, 1.0)
            material = DiffuseMaterial(default_color)

        super().__init__(
            name,
            translation,
            rotation,
            scale,
            material,
            enabled,
            viewport_mask,
        )

        # Add properties
        self.positions = self.add_buffer_property(positions, np.float32, (-1, 3), name="positions").use_gpu(
            BufferUsageFlags.VERTEX
            | BufferUsageFlags.SHADER_DEVICE_ADDRESS
            | BufferUsageFlags.ACCELERATION_STRUCTURE_INPUT,
            PipelineStageFlags.VERTEX_INPUT,
        )
        self.normals = (
            self.add_buffer_property(normals, np.float32, (-1, 3), name="normals").use_gpu(
                BufferUsageFlags.VERTEX | BufferUsageFlags.SHADER_DEVICE_ADDRESS, PipelineStageFlags.VERTEX_INPUT
            )
            if normals is not None
            else None
        )
        self.tangents = (
            self.add_buffer_property(tangents, np.float32, (-1, 3), name="tangents").use_gpu(
                BufferUsageFlags.VERTEX | BufferUsageFlags.SHADER_DEVICE_ADDRESS, PipelineStageFlags.VERTEX_INPUT
            )
            if tangents is not None
            else None
        )
        self.uvs = (
            self.add_buffer_property(uvs, np.float32, (-1, 2), name="uvs").use_gpu(
                BufferUsageFlags.VERTEX | BufferUsageFlags.SHADER_DEVICE_ADDRESS, PipelineStageFlags.VERTEX_INPUT
            )
            if uvs is not None
            else None
        )
        self.indices = (
            self.add_buffer_property(indices, np.uint32, (-1,), name="indices").use_gpu(
                BufferUsageFlags.INDEX
                | BufferUsageFlags.SHADER_DEVICE_ADDRESS
                | BufferUsageFlags.ACCELERATION_STRUCTURE_INPUT,
                PipelineStageFlags.VERTEX_INPUT,
            )
            if indices is not None
            else None
        )
        self.vertex_colors = (
            self.add_buffer_property(vertex_colors, np.uint32, (-1,), name="vertex_colors").use_gpu(
                BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
            )
            if vertex_colors is not None
            else None
        )
        self.instance_positions = (
            self.add_buffer_property(instance_positions, np.float32, (-1, 3), name="instance_positions").use_gpu(
                BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
            )
            if instance_positions is not None
            else None
        )

    def create(self, r: Renderer) -> None:
        assert self.material is not None
        defines: List[Tuple[str, str]] = []
        defines.extend(self.material.shader_defines)

        depth_defines: List[Tuple[str, str]] = []
        vertex_binding_index = 1
        vertex_bindings = [
            VertexBinding(0, 12, VertexInputRate.VERTEX),
        ]
        vertex_attributes = [
            VertexAttribute(0, 0, Format.R32G32B32_SFLOAT),
        ]

        depth_vertex_binding_index = 1
        depth_vertex_bindings = [
            VertexBinding(0, 12, VertexInputRate.VERTEX),
        ]
        depth_vertex_attributes = [
            VertexAttribute(0, 0, Format.R32G32B32_SFLOAT),
        ]

        if self.normals is not None:
            defines.append(("VERTEX_NORMALS", str(vertex_binding_index)))
            vertex_bindings.append(VertexBinding(vertex_binding_index, 12, VertexInputRate.VERTEX))
            vertex_attributes.append(
                VertexAttribute(vertex_binding_index, vertex_binding_index, Format.R32G32B32_SFLOAT)
            )
            vertex_binding_index += 1

        if self.tangents is not None:
            defines.append(("VERTEX_TANGENTS", str(vertex_binding_index)))
            vertex_bindings.append(VertexBinding(vertex_binding_index, 12, VertexInputRate.VERTEX))
            vertex_attributes.append(
                VertexAttribute(vertex_binding_index, vertex_binding_index, Format.R32G32B32_SFLOAT)
            )
            vertex_binding_index += 1

        if self.uvs is not None:
            defines.append(("VERTEX_UVS", str(vertex_binding_index)))
            vertex_bindings.append(VertexBinding(vertex_binding_index, 8, VertexInputRate.VERTEX))
            vertex_attributes.append(VertexAttribute(vertex_binding_index, vertex_binding_index, Format.R32G32_SFLOAT))
            vertex_binding_index += 1

        if self.vertex_colors is not None:
            defines.append(("VERTEX_COLORS", str(vertex_binding_index)))
            vertex_bindings.append(VertexBinding(vertex_binding_index, 4, VertexInputRate.VERTEX))
            vertex_attributes.append(VertexAttribute(vertex_binding_index, vertex_binding_index, Format.R32_UINT))
            vertex_binding_index += 1

        if self.instance_positions is not None:
            defines.append(("INSTANCE_POSITIONS", str(vertex_binding_index)))
            vertex_bindings.append(VertexBinding(vertex_binding_index, 12, VertexInputRate.INSTANCE))
            vertex_attributes.append(
                VertexAttribute(vertex_binding_index, vertex_binding_index, Format.R32G32B32_SFLOAT)
            )
            vertex_binding_index += 1

            depth_defines.append(("INSTANCE_POSITIONS", str(depth_vertex_binding_index)))
            depth_vertex_bindings.append(VertexBinding(depth_vertex_binding_index, 12, VertexInputRate.INSTANCE))
            depth_vertex_attributes.append(
                VertexAttribute(depth_vertex_binding_index, depth_vertex_binding_index, Format.R32G32B32_SFLOAT)
            )
            depth_vertex_binding_index += 1

        vert = r.compile_builtin_shader("3d/mesh.slang", "vertex_main", defines=defines)
        frag = r.compile_builtin_shader("3d/mesh.slang", "pixel_main", defines=defines)

        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=vertex_bindings,
            vertex_attributes=vertex_attributes,
            rasterization=Rasterization(
                polygon_mode=self.polygon_mode, cull_mode=self.cull_mode, front_face=self.front_face
            ),
            input_assembly=InputAssembly(self.primitive_topology),
            samples=r.msaa_samples,
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
                self.material.descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

        depth_vert = r.compile_builtin_shader("3d/mesh_depth.slang", "vertex_main", defines=depth_defines)
        self.depth_pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, depth_vert.code), Stage.VERTEX),
            ],
            vertex_bindings=depth_vertex_bindings,
            vertex_attributes=depth_vertex_attributes,
            rasterization=Rasterization(
                polygon_mode=self.polygon_mode,
                cull_mode=cull_mode_opposite_face(self.cull_mode),
                front_face=self.front_face,
            ),
            input_assembly=InputAssembly(self.primitive_topology),
            attachments=[],
            depth=Depth(r.shadow_map_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_depth_descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype["transform"].itemsize)],
        )

    def append_acceleration_structure_instances(
        self, instances: List[AccelerationStructureInstanceInfo], material_index: int
    ) -> None:
        # NOTE: acceleration structure build only supports triangle list. Other triangle topologies could
        # be supported if we were to manually expand the index buffer.
        if self.primitive_topology != PrimitiveTopology.TRIANGLE_LIST:
            return

        def buffer_address_or_null(p: Optional[BufferProperty]) -> int:
            if p is None:
                return 0
            view = p.get_current_gpu()
            return view.buffer.address + view.offset

        positions = self.positions.get_current_gpu()
        positions_count = positions.size // 12

        indices_addr = 0
        if self.indices is not None:
            indices = self.indices.get_current_gpu()
            indices_addr = indices.buffer.address + indices.offset
            primitive_count = indices.size // 12
        else:
            primitive_count = positions_count // 3

        # TODO: proper instances without duplicating blases
        if self.instance_positions is not None:
            instance_positions = self.instance_positions.get_current()
            for pos in instance_positions:
                transform = self.constants["transform"][0].copy()
                transform[:3, 3] += pos
                instances.append(
                    AccelerationStructureInstanceInfo(
                        transform,
                        self.constants["normal_matrix"],
                        positions_count,
                        positions.buffer.address + positions.offset,
                        buffer_address_or_null(self.normals),
                        buffer_address_or_null(self.tangents),
                        buffer_address_or_null(self.uvs),
                        primitive_count,
                        indices_addr,
                        material_index,
                        self.viewport_mask if self.viewport_mask is not None else 0xFF,
                    )
                )
        else:
            instances.append(
                AccelerationStructureInstanceInfo(
                    self.constants["transform"],
                    self.constants["normal_matrix"],
                    positions_count,
                    positions.buffer.address + positions.offset,
                    buffer_address_or_null(self.normals),
                    buffer_address_or_null(self.tangents),
                    buffer_address_or_null(self.uvs),
                    primitive_count,
                    indices_addr,
                    material_index,
                    self.viewport_mask if self.viewport_mask is not None else 0xFF,
                )
            )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"] = mat4x3(self.current_transform_matrix)
        self.constants["normal_matrix"][:, :, :3] = transpose(inverse(mat3(self.current_transform_matrix)))

    def render_depth(self, r: Renderer, frame: RendererFrame, scene_descriptor_set: DescriptorSet) -> None:
        positions = self.positions.get_current_gpu()
        vertex_buffers = [
            positions.buffer_and_offset(),
        ]
        num_instances = 1
        if self.instance_positions is not None:
            vertex_buffers.append(self.instance_positions.get_current_gpu().buffer_and_offset())
            instance_positions_buf = self.instance_positions.get_current_gpu()
            vertex_buffers.append(instance_positions_buf.buffer_and_offset())
            num_instances = instance_positions_buf.size // 12

        indices = self.indices.get_current_gpu() if self.indices is not None else None
        frame.cmd.bind_graphics_pipeline(
            self.depth_pipeline,
            vertex_buffers=vertex_buffers,
            index_buffer=indices.buffer_and_offset() if indices is not None else None,
            descriptor_sets=[
                scene_descriptor_set,
            ],
            push_constants=self.constants["transform"].tobytes(),
        )

        if indices is not None:
            frame.cmd.draw_indexed(indices.size // 4, num_instances)
        else:
            frame.cmd.draw(positions.size // 12, num_instances)

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        assert self.material is not None

        positions = self.positions.get_current_gpu()
        vertex_buffers = [
            positions.buffer_and_offset(),
        ]
        num_instances = 1
        if self.normals is not None:
            vertex_buffers.append(self.normals.get_current_gpu().buffer_and_offset())
        if self.tangents is not None:
            vertex_buffers.append(self.tangents.get_current_gpu().buffer_and_offset())
        if self.uvs is not None:
            vertex_buffers.append(self.uvs.get_current_gpu().buffer_and_offset())
        if self.vertex_colors is not None:
            vertex_buffers.append(self.vertex_colors.get_current_gpu().buffer_and_offset())
        if self.instance_positions is not None:
            instance_positions_buf = self.instance_positions.get_current_gpu()
            vertex_buffers.append(instance_positions_buf.buffer_and_offset())
            num_instances = instance_positions_buf.size // 12

        indices = self.indices.get_current_gpu() if self.indices is not None else None
        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=vertex_buffers,
            index_buffer=indices.buffer_and_offset() if indices is not None else None,
            descriptor_sets=[
                scene_descriptor_set,
                self.material.descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )

        if indices is not None:
            frame.cmd.draw_indexed(indices.size // 4, num_instances)
        else:
            frame.cmd.draw(positions.size // 12, num_instances)


class MarchingCubesMesh(Object3D):
    def __init__(
        self,
        sdf: ImageProperty,
        size: Tuple[float, float, float],
        level: float = 0.0,
        invert_normals: bool = False,
        max_vertices: Optional[int] = None,
        max_indices: Optional[int] = None,
        material: Optional[Material] = None,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
                "normal_matrix": (np.dtype((np.float32, (3, 4))), 48),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)

        if material is None:
            material = DiffuseMaterial((0.5, 0.5, 0.5))

        super().__init__(
            name,
            translation,
            rotation,
            scale,
            material,
            enabled,
            viewport_mask,
        )

        self.size = size
        self.level = level
        self.invert_normals = invert_normals

        # Add properties
        self.sdf = self.add_image_property(sdf, is_3d=True, name="sdf").use_gpu(
            ImageUsageFlags.STORAGE,
            ImageLayout.GENERAL,
            PipelineStageFlags.COMPUTE_SHADER,
        )

        if self.sdf.depth is None:
            raise RuntimeError("sdf must be a 3D image")

        width, height, depth = self.sdf.width, self.sdf.height, self.sdf.depth
        self.blocks_x = div_ceil(width, BLOCK_SIZE_X - 1)
        self.blocks_y = div_ceil(height, BLOCK_SIZE_Y - 1)
        self.blocks_z = div_ceil(depth, BLOCK_SIZE_Z - 1)
        self.total_blocks = self.blocks_x * self.blocks_y * self.blocks_z

        # Compute reasonable upper bounds on number of vertices and indices based on sdf area
        area = (width * height + width * depth + height * depth) * 2
        if max_vertices is None:
            max_vertices = max(align_up(area * 2, 64), 1024)
        if max_indices is None:
            max_indices = max(align_up(area * 8, 64), 4096)

        self.max_vertices = max_vertices
        self.max_indices = max_indices

        self._need_marching = True

        self.sdf.update_callbacks.append(lambda _: self.invalidate_surface())

    def create(self, r: Renderer) -> None:
        assert self.material is not None
        defines: List[Tuple[str, str]] = [
            ("VERTEX_NORMALS", "1"),
        ]
        defines.extend(self.material.shader_defines)

        depth_defines: List[Tuple[str, str]] = []
        vertex_bindings = [
            VertexBinding(0, 12, VertexInputRate.VERTEX),
            VertexBinding(1, 12, VertexInputRate.VERTEX),
        ]
        vertex_attributes = [
            VertexAttribute(0, 0, Format.R32G32B32_SFLOAT),
            VertexAttribute(1, 1, Format.R32G32B32_SFLOAT),
        ]

        depth_vertex_bindings = [
            VertexBinding(0, 12, VertexInputRate.VERTEX),
        ]
        depth_vertex_attributes = [
            VertexAttribute(0, 0, Format.R32G32B32_SFLOAT),
        ]

        vert = r.compile_builtin_shader("3d/mesh.slang", "vertex_main", defines=defines)
        frag = r.compile_builtin_shader("3d/mesh.slang", "pixel_main", defines=defines)

        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=vertex_bindings,
            vertex_attributes=vertex_attributes,
            rasterization=Rasterization(cull_mode=CullMode.NONE),
            input_assembly=InputAssembly(PrimitiveTopology.TRIANGLE_LIST),
            samples=r.msaa_samples,
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
                self.material.descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

        depth_vert = r.compile_builtin_shader("3d/mesh_depth.slang", "vertex_main", defines=depth_defines)
        self.depth_pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, depth_vert.code), Stage.VERTEX),
            ],
            vertex_bindings=depth_vertex_bindings,
            vertex_attributes=depth_vertex_attributes,
            rasterization=Rasterization(cull_mode=CullMode.NONE),
            input_assembly=InputAssembly(PrimitiveTopology.TRIANGLE_LIST),
            attachments=[],
            depth=Depth(r.shadow_map_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_depth_descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype["transform"].itemsize)],
        )

        self.blocks_buf = Buffer(
            r.device,
            self.total_blocks * 4,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC),
            AllocType.DEVICE,
            name="marching-cubes-mesh-blocks",
        )
        self.valid_blocks_buf = Buffer(
            r.device,
            self.total_blocks * 4,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC),
            AllocType.DEVICE,
            name="marching-cubes-mesh-valid-blocks",
        )

        acceleration_structure_flags = 0
        address_flags = 0
        if r.supports_path_tracing:
            acceleration_structure_flags |= (
                BufferUsageFlags.SHADER_DEVICE_ADDRESS | BufferUsageFlags.ACCELERATION_STRUCTURE_INPUT
            )
            address_flags |= BufferUsageFlags.SHADER_DEVICE_ADDRESS

        self.positions_buf = Buffer(
            r.device,
            self.max_vertices * 12,
            (
                BufferUsageFlags.STORAGE
                | BufferUsageFlags.TRANSFER_SRC
                | BufferUsageFlags.VERTEX
                | acceleration_structure_flags
            ),
            AllocType.DEVICE,
            name="marching-cubes-mesh-positions",
        )
        self.normals_buf = Buffer(
            r.device,
            self.max_vertices * 12,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC | BufferUsageFlags.VERTEX | address_flags),
            AllocType.DEVICE,
            name="marching-cubes-mesh-normals",
        )
        self.indices_buf = Buffer(
            r.device,
            self.max_indices * 4,
            (
                BufferUsageFlags.STORAGE
                | BufferUsageFlags.TRANSFER_SRC
                | BufferUsageFlags.INDEX
                | acceleration_structure_flags
            ),
            AllocType.DEVICE,
            name="marching-cubes-mesh-indices",
        )

        self.marching_cubes_pipeline = MarchingCubesPipeline(r)
        self.marching_cubes_instance = self.marching_cubes_pipeline.alloc_instance(r)

        instance = self.marching_cubes_instance

        for check_set, compact_set, march_set in zip(
            instance.check_descriptor_sets, instance.compact_descriptor_sets, instance.march_descriptor_sets
        ):
            check_set.write_buffer(self.blocks_buf, DescriptorType.STORAGE_BUFFER, 1)

            compact_set.write_buffer(self.blocks_buf, DescriptorType.STORAGE_BUFFER, 0)
            compact_set.write_buffer(self.valid_blocks_buf, DescriptorType.STORAGE_BUFFER, 2)

            march_set.write_buffer(self.valid_blocks_buf, DescriptorType.STORAGE_BUFFER, 1)
            march_set.write_buffer(self.positions_buf, DescriptorType.STORAGE_BUFFER, 3)
            march_set.write_buffer(self.normals_buf, DescriptorType.STORAGE_BUFFER, 4)
            march_set.write_buffer(self.indices_buf, DescriptorType.STORAGE_BUFFER, 5)

    def append_acceleration_structure_instances(
        self, instances: List[AccelerationStructureInstanceInfo], material_index: int
    ) -> None:
        # Assume marching cube results are available for readback.
        mc_instance = self.marching_cubes_instance

        vertices_counter_readback_buf = (
            mc_instance.vertices_counter_readback_buf
            if mc_instance.vertices_counter_readback_buf is not None
            else mc_instance.vertices_counter_buf
        )
        vertices_arr = np.frombuffer(vertices_counter_readback_buf.data, np.uint32)
        positions_count = vertices_arr[0]

        indices_counter_readback_buf = (
            mc_instance.indices_counter_readback_buf
            if mc_instance.indices_counter_readback_buf is not None
            else mc_instance.indices_counter_buf
        )
        indices_arr = np.frombuffer(indices_counter_readback_buf.data, np.uint32)
        primitive_count = indices_arr[0] // 3

        instances.append(
            AccelerationStructureInstanceInfo(
                self.constants["transform"],
                self.constants["normal_matrix"],
                positions_count,
                self.positions_buf.address,
                self.normals_buf.address,
                0,
                0,
                primitive_count,
                self.indices_buf.address,
                material_index,
                self.viewport_mask if self.viewport_mask is not None else 0xFF,
            )
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"] = mat4x3(self.current_transform_matrix)
        self.constants["normal_matrix"][:, :, :3] = transpose(inverse(mat3(self.current_transform_matrix)))

    def _march(self, r: Renderer, frame: RendererFrame) -> None:
        sdf_view = self.sdf.get_current_gpu().view()
        self.marching_cubes_instance.check_descriptor_sets.get_current().write_image(
            sdf_view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0
        )
        self.marching_cubes_instance.march_descriptor_sets.get_current().write_image(
            sdf_view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0
        )

        assert self.sdf.depth is not None
        self.marching_cubes_pipeline.run(
            frame.cmd,
            self.sdf.width,
            self.sdf.height,
            self.sdf.depth,
            self.blocks_x,
            self.blocks_y,
            self.blocks_z,
            self.total_blocks,
            self.max_vertices,
            self.max_indices,
            self.marching_cubes_instance,
            self.size,
            self.level,
            self.invert_normals,
        )

    def invalidate_surface(self) -> None:
        self._need_marching = True

    def post_upload(self, r: Renderer, frame: RendererFrame) -> None:
        if self._need_marching:
            self._march(r, frame)
            self._need_marching = False

    def render_depth(self, r: Renderer, frame: RendererFrame, scene_descriptor_set: DescriptorSet) -> None:
        frame.cmd.bind_graphics_pipeline(
            self.depth_pipeline,
            vertex_buffers=[self.positions_buf],
            index_buffer=self.indices_buf,
            descriptor_sets=[
                scene_descriptor_set,
            ],
            push_constants=self.constants["transform"].tobytes(),
        )

        frame.cmd.draw_indexed_indirect(self.marching_cubes_instance.indices_counter_buf, 0, 1, 20)

    def render(
        self, r: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        assert self.material is not None

        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=[self.positions_buf, self.normals_buf],
            index_buffer=self.indices_buf,
            descriptor_sets=[
                scene_descriptor_set,
                self.material.descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )

        frame.cmd.draw_indexed_indirect(self.marching_cubes_instance.indices_counter_buf, 0, 1, 20)


class Sphere(Mesh):
    def __init__(
        self,
        radius: float = 0.1,
        color: int = 0xFFCCCCCC,
        rings: int = 16,
        sectors: int = 32,
        instance_positions: Optional[BufferProperty] = None,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        v, n, f = create_sphere(radius, rings, sectors)
        c = np.full(v.shape[0], color, np.uint32)

        super().__init__(
            v,  # type: ignore
            f,  # type: ignore
            n,  # type: ignore
            vertex_colors=c,  # type: ignore
            instance_positions=instance_positions,
            name=name,
            translation=translation,
            rotation=rotation,
            scale=scale,
            enabled=enabled,
            viewport_mask=viewport_mask,
        )


class AxisGizmo(Mesh):
    def __init__(
        self,
        sphere_radius: float = 0.01,
        axis_radius: float = 0.005,
        axis_length: float = 0.1,
        sphere_color: int = 0xFFCCCCCC,
        x_axis_color: int = 0xFF0000FF,
        y_axis_color: int = 0xFF00FF00,
        z_axis_color: int = 0xFFFF0000,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        sphere_v, sphere_n, sphere_f = create_sphere(sphere_radius, 8, 16)
        sphere_c = np.full(sphere_v.shape[0], sphere_color, np.uint32)

        z_cylinder_v, z_cylinder_n, cylinder_f = create_arrow(
            axis_radius, axis_length * 0.8, axis_radius * 2.0, axis_length * 0.2, 16
        )
        z_cylinder_c = np.full(z_cylinder_v.shape[0], z_axis_color, np.uint32)

        x_cylinder_v, x_cylinder_n = transform_mesh(rotate(np.pi * 0.5, vec3(0, 1, 0)), z_cylinder_v, z_cylinder_n)
        x_cylinder_c = np.full(x_cylinder_v.shape[0], x_axis_color, np.uint32)

        y_cylinder_v, y_cylinder_n = transform_mesh(rotate(np.pi * 0.5, vec3(0, 0, 1)), x_cylinder_v, x_cylinder_n)
        y_cylinder_c = np.full(y_cylinder_v.shape[0], y_axis_color, np.uint32)
        (mesh_v, mesh_n, mesh_c), mesh_f = concatenate_meshes(
            [
                (sphere_v, sphere_n, sphere_c),
                (x_cylinder_v, x_cylinder_n, x_cylinder_c),
                (y_cylinder_v, y_cylinder_n, y_cylinder_c),
                (z_cylinder_v, z_cylinder_n, z_cylinder_c),
            ],
            [sphere_f, cylinder_f, cylinder_f, cylinder_f],
        )

        super().__init__(
            mesh_v,  # type: ignore
            mesh_f,  # type: ignore
            mesh_n,  # type: ignore
            vertex_colors=mesh_c,  # type: ignore
            name=name,
            translation=translation,
            rotation=rotation,
            scale=scale,
            enabled=enabled,
            viewport_mask=viewport_mask,
        )


class Image(Mesh):
    def __init__(
        self,
        image: ImageProperty,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        material = ColorMaterial(as_image_property(image))
        positions = np.array(
            [
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                1.0,
                0.0,
                0.0,
                1.0,
                1.0,
                0.0,
            ]
        ).reshape((4, 3))
        uvs = np.array(
            [
                0.0,
                0.0,
                0.0,
                1.0,
                1.0,
                0.0,
                1.0,
                1.0,
            ]
        ).reshape((4, 2))
        super().__init__(
            positions,  # type: ignore
            uvs=uvs,  # type: ignore
            primitive_topology=PrimitiveTopology.TRIANGLE_STRIP,
            material=material,
            name=name,
            translation=translation,
            rotation=rotation,
            scale=scale,
            enabled=enabled,
            viewport_mask=viewport_mask,
        )


class GridType(Enum):
    XY_PLANE = 0
    YZ_PLANE = 1
    XZ_PLANE = 2


class Grid(Object3D):
    def __init__(
        self,
        size: Tuple[float, float],
        grid_type: GridType,
        major_line_color: Tuple[float, float, float, float],
        minor_line_color: Tuple[float, float, float, float],
        base_color: Tuple[float, float, float, float],
        grid_scale: float = 0.1,
        major_grid_div: float = 10.0,
        axis_line_width: float = 0.08,
        major_line_width: float = 0.04,
        minor_line_width: float = 0.01,
        pos_axis_color_scale: float = 1.0,
        neg_axis_color_scale: float = 0.5,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)

        self.grid_constants_dtype = np.dtype(
            {
                "major_line_color": (np.dtype((np.float32, 4)), 0),
                "minor_line_color": (np.dtype((np.float32, 4)), 16),
                "base_color": (np.dtype((np.float32, 4)), 32),
                "size": (np.dtype((np.float32, 2)), 48),
                "grid_type": (np.uint32, 56),
                "inv_grid_scale": (np.float32, 60),
                "major_grid_div": (np.float32, 64),
                "axis_line_width": (np.float32, 68),
                "major_line_width": (np.float32, 72),
                "minor_line_width": (np.float32, 76),
                "pos_axis_color_scale": (np.float32, 80),
                "neg_axis_color_scale": (np.float32, 84),
            }
        )  # type: ignore
        self.grid_constants = np.zeros((1,), self.constants_dtype)

        self.grid_constants = np.zeros((1,), self.grid_constants_dtype)
        self.grid_constants["major_line_color"] = major_line_color
        self.grid_constants["minor_line_color"] = minor_line_color
        self.grid_constants["base_color"] = base_color
        self.grid_constants["size"] = size
        self.grid_constants["grid_type"] = grid_type.value
        self.grid_constants["inv_grid_scale"] = 1.0 / grid_scale
        self.grid_constants["major_grid_div"] = major_grid_div
        self.grid_constants["axis_line_width"] = axis_line_width
        self.grid_constants["major_line_width"] = major_line_width
        self.grid_constants["minor_line_width"] = minor_line_width
        self.grid_constants["pos_axis_color_scale"] = pos_axis_color_scale
        self.grid_constants["neg_axis_color_scale"] = neg_axis_color_scale

        self.is_transparent = base_color[3] < 1.0

        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)

    @classmethod
    def white(cls, size: Tuple[float, float], grid_type: GridType, **kwargs: Any) -> "Grid":
        return cls(size, grid_type, (0, 0, 0, 1), (0, 0, 0, 1), (1, 1, 1, 1), **kwargs)

    @classmethod
    def black(cls, size: Tuple[float, float], grid_type: GridType, **kwargs: Any) -> "Grid":
        return cls(size, grid_type, (1, 1, 1, 1), (1, 1, 1, 1), (0, 0, 0, 1), **kwargs)

    @classmethod
    def transparent_white_lines(cls, size: Tuple[float, float], grid_type: GridType, **kwargs: Any) -> "Grid":
        return cls(size, grid_type, (1, 1, 1, 1), (1, 1, 1, 1), (1, 1, 1, 0), **kwargs)

    @classmethod
    def transparent_black_lines(cls, size: Tuple[float, float], grid_type: GridType, **kwargs: Any) -> "Grid":
        return cls(size, grid_type, (0, 0, 0, 1), (0, 0, 0, 1), (0, 0, 0, 0), **kwargs)

    @classmethod
    def dark_gray_white_lines(cls, size: Tuple[float, float], grid_type: GridType, **kwargs: Any) -> "Grid":
        return cls(size, grid_type, (1, 1, 1, 1), (1, 1, 1, 1), (0.1, 0.1, 0.1, 1), **kwargs)

    @classmethod
    def light_gray_black_lines(cls, size: Tuple[float, float], grid_type: GridType, **kwargs: Any) -> "Grid":
        return cls(size, grid_type, (0, 0, 0, 1), (0, 0, 0, 1), (0.9, 0.9, 0.9, 1), **kwargs)

    def create(self, r: Renderer) -> None:
        vert = r.compile_builtin_shader("3d/grid.slang", "vertex_main")
        frag = r.compile_builtin_shader("3d/grid.slang", "pixel_main")

        self.grid_constants_buf = Buffer.from_data(
            r.device,
            view_bytes(self.grid_constants),
            BufferUsageFlags.TRANSFER_DST | BufferUsageFlags.UNIFORM,
            AllocType.DEVICE_MAPPED_WITH_FALLBACK,
        )
        self.descriptor_set_layout, self.descriptor_pool, self.descriptor_set = create_descriptor_layout_pool_and_set(
            r.device,
            [DescriptorSetBinding(1, DescriptorType.UNIFORM_BUFFER)],
        )
        self.descriptor_set.write_buffer(self.grid_constants_buf, DescriptorType.UNIFORM_BUFFER, 0)

        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            rasterization=Rasterization(cull_mode=CullMode.NONE),
            input_assembly=InputAssembly(PrimitiveTopology.TRIANGLE_STRIP),
            samples=r.msaa_samples,
            attachments=[
                Attachment(
                    format=r.output_format,
                    blend_enable=True,
                    src_color_blend_factor=BlendFactor.SRC_ALPHA,
                    dst_color_blend_factor=BlendFactor.ONE_MINUS_SRC_ALPHA,
                    color_blend_op=BlendOp.ADD,
                    src_alpha_blend_factor=BlendFactor.ONE,
                    dst_alpha_blend_factor=BlendFactor.ONE_MINUS_SRC_ALPHA,
                    alpha_blend_op=BlendOp.ADD,
                )
            ],
            depth=Depth(r.depth_format, True, not self.is_transparent, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
                self.descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"] = mat4x3(self.current_transform_matrix)

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            descriptor_sets=[
                scene_descriptor_set,
                self.descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )
        frame.cmd.draw(4)


class Voxels(Object3D):
    def __init__(
        self,
        positions: BufferProperty,
        size: float,
        colors: Optional[BufferProperty] = None,
        uniform_color: Optional[int] = None,
        colormap: Optional[Colormap] = None,
        border_size: float = 0.1,
        border_factor: float = 0.5,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        if uniform_color is None and colors is None and colormap is None:
            uniform_color = 0xFFCCCCCC

        if uniform_color is not None:
            if colormap is not None or colors is not None:
                raise ValueError("Only one of uniform_color, colormap and colors can be not None")
            self.constants_dtype = np.dtype(
                {
                    "transform": (np.dtype((np.float32, (3, 4))), 0),
                    "object_camera_position": (np.dtype((np.float32, (3,))), 48),
                    "size": (np.float32, 60),
                    "border_size": (np.float32, 64),
                    "border_factor": (np.float32, 68),
                    "color": (np.uint32, 72),
                }
            )  # type: ignore
        elif colors is not None:
            if colormap is not None:
                raise ValueError("Only one of uniform_color, colormap and colors can be not None")
            self.constants_dtype = np.dtype(
                {
                    "transform": (np.dtype((np.float32, (3, 4))), 0),
                    "object_camera_position": (np.dtype((np.float32, (3,))), 48),
                    "size": (np.float32, 60),
                    "border_size": (np.float32, 64),
                    "border_factor": (np.float32, 68),
                }
            )  # type: ignore
        elif colormap is not None:
            self.constants_dtype = np.dtype(
                {
                    "transform": (np.dtype((np.float32, (3, 4))), 0),
                    "object_camera_position": (np.dtype((np.float32, (3,))), 48),
                    "size": (np.float32, 60),
                    "border_size": (np.float32, 64),
                    "border_factor": (np.float32, 68),
                    "colormap_measure": (np.uint32, 72),
                    "range_min": (np.float32, 76),
                    "point_or_normal": (np.dtype((np.float32, (3,))), 80),
                    "range_inv_delta": (np.float32, 92),
                }
            )  # type: ignore

        self.constants = np.zeros((1,), self.constants_dtype)
        self.size = size
        self.border_size = border_size
        self.border_factor = border_factor

        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)
        self.positions = self.add_buffer_property(positions, np.float32, (-1, 3), name="positions").use_gpu(
            BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
        )
        self.colors = (
            self.add_buffer_property(colors, np.uint32, (-1,), name="colors").use_gpu(
                BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
            )
            if colors is not None
            else None
        )
        self.colormap = colormap
        self.uniform_color = uniform_color

    def create(self, r: Renderer) -> None:
        vertex_bindings = [VertexBinding(0, 12, VertexInputRate.INSTANCE)]
        vertex_attributes = [VertexAttribute(0, 0, Format.R32G32B32_SFLOAT)]

        defines = []
        if self.uniform_color is not None:
            defines.append(("UNIFORM_COLOR", ""))
        elif self.colors is not None:
            defines.append(("VERTEX_COLORS", ""))
            vertex_bindings.append(VertexBinding(1, 4, VertexInputRate.INSTANCE))
            vertex_attributes.append(VertexAttribute(1, 1, Format.R32_UINT))
        elif self.colormap is not None:
            defines.append(("COLORMAP", ""))

        vert = r.compile_builtin_shader("3d/voxels.slang", "vertex_main", defines=defines)
        frag = r.compile_builtin_shader("3d/voxels.slang", "pixel_main", defines=defines)

        indices = np.array(
            [
                [0, 1, 3],
                [3, 0, 2],
                [0, 4, 5],
                [1, 0, 5],
                [0, 2, 4],
                [2, 4, 6],
            ],
            np.uint32,
        )
        self.indices = Buffer.from_data(
            r.device,
            view_bytes(indices),
            BufferUsageFlags.INDEX | BufferUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            name="voxel-indices",
        )

        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=vertex_bindings,
            vertex_attributes=vertex_attributes,
            input_assembly=InputAssembly(PrimitiveTopology.TRIANGLE_LIST),
            samples=r.msaa_samples,
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"] = mat4x3(self.current_transform_matrix)

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        self.constants["object_camera_position"] = vec3(
            inverse(self.current_transform_matrix) * vec4(viewport.camera.position(), 1.0)  # type: ignore
        )
        self.constants["size"] = self.size
        self.constants["border_size"] = self.border_size
        self.constants["border_factor"] = self.border_factor
        if self.uniform_color is not None:
            self.constants["color"] = self.uniform_color
        elif self.colormap is not None:
            if isinstance(self.colormap, ColormapDistanceToPoint):
                measure = 0
                self.constants["point_or_normal"][:3] = self.colormap.point
            elif isinstance(self.colormap, ColormapDistanceToPlane):
                measure = 1
                self.constants["point_or_normal"][:3] = self.colormap.normal
            else:
                raise ValueError("colormap must be of type ColormapDistanceToPlane or ColormapDistanceToPoint")
            self.constants["colormap_measure"] = (measure << 16) | self.colormap.color.value
            self.constants["range_min"] = self.colormap.range_min
            self.constants["range_inv_delta"] = 1.0 / (self.colormap.range_max - self.colormap.range_min)

        positions = self.positions.get_current_gpu()

        vertex_buffers = [positions.buffer_and_offset()]
        if self.colors is not None:
            vertex_buffers.append(self.colors.get_current_gpu().buffer_and_offset())

        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=vertex_buffers,
            index_buffer=self.indices,
            descriptor_sets=[
                scene_descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )
        frame.cmd.draw_indexed(18, positions.size // 12)


# NOTE: must be kept in sync with RenderFlags in shaders/3d/gaussian_splatting/shaderio.h
class GaussianSplatsRenderFlags(IntFlag):
    NONE = (0,)
    DISABLE_OPACITY = (1 << 0,)
    SHOW_SPHERICAL_HARMONICS_ONLY = (1 << 1,)
    ORTHOGRAPHIC_MODE = (1 << 2,)
    POINT_CLOUD_MODE = (1 << 3,)


class GaussianSplats(Object3D):
    DISTANCE_COMPUTE_WORKGROUP_SIZE = 256

    FRUSTUM_CULLING_AT_NONE = 0
    FRUSTUM_CULLING_AT_DIST = 1
    FRUSTUM_CULLING_AT_RASTER = 2

    FORMAT_FLOAT32 = 0
    FORMAT_FLOAT16 = 1
    FORMAT_UINT8 = 2

    def __init__(
        self,
        positions: BufferProperty,
        colors: BufferProperty,
        spherical_harmonics: BufferProperty,
        covariances: BufferProperty,
        flags: GaussianSplatsRenderFlags = GaussianSplatsRenderFlags.NONE,
        cull_at_dist: bool = True,
        deterministic_cull_at_dist: bool = True,
        mip_splatting_antialiasing: bool = False,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
                "inverse_transform": (np.dtype((np.float32, (3, 4))), 48),
                "splat_count": (np.uint32, 96),
                "alpha_cull_threshold": (np.float32, 100),
                "frustum_dilation": (np.float32, 104),
                "splat_scale": (np.float32, 108),
                "flags": (np.uint32, 112),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)

        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)
        self.positions = self.add_buffer_property(positions, np.float32, (-1, 3), name="positions").use_gpu(
            BufferUsageFlags.STORAGE,
            PipelineStageFlags.COMPUTE_SHADER,
        )
        self.colors = self.add_buffer_property(colors, np.float32, (-1, 4), name="colors").use_gpu(
            BufferUsageFlags.STORAGE,
            PipelineStageFlags.COMPUTE_SHADER,
        )
        self.spherical_harmonics = self.add_buffer_property(
            spherical_harmonics, None, (-1, -1), name="spherical-harmonics"
        ).use_gpu(
            BufferUsageFlags.STORAGE,
            PipelineStageFlags.COMPUTE_SHADER,
        )
        if self.spherical_harmonics.dtype == np.uint8:
            self.sh_format = self.FORMAT_UINT8
        elif self.spherical_harmonics.dtype == np.float16:
            self.sh_format = self.FORMAT_FLOAT16
        elif self.spherical_harmonics.dtype == np.float32:
            self.sh_format = self.FORMAT_FLOAT32
        else:
            raise TypeError(
                f"Spherical harmonics must be of dtype uint8, float16 or float32. Got {self.spherical_harmonics.dtype}."
            )
        self.covariances = self.add_buffer_property(covariances, np.float32, (-1, 6), name="covariances").use_gpu(
            BufferUsageFlags.STORAGE,
            PipelineStageFlags.COMPUTE_SHADER,
        )
        self.mip_splatting_antialiasing = mip_splatting_antialiasing
        self.flags = flags
        self.cull_at_dist = cull_at_dist
        self.deterministic_cull_at_dist = deterministic_cull_at_dist
        self.alpha_cull_threshold = 1.0 / 255.0
        self.frustum_dilation = 0.2
        self.splat_scale = 1.0

    def create(self, r: Renderer) -> None:
        if self.sh_format == self.FORMAT_UINT8 and (r.device.features & DeviceFeatures.STORAGE_8BIT) == 0:
            raise RuntimeError("Device does not support 8bit storage")
        if self.sh_format == self.FORMAT_FLOAT16 and (r.device.features & DeviceFeatures.STORAGE_16BIT) == 0:
            raise RuntimeError("Device does not support 16bit storage")

        self.use_barycentric = (r.device.features & DeviceFeatures.FRAGMENT_SHADER_BARYCENTRIC) != 0
        self.use_mesh_shader = (r.device.features & DeviceFeatures.MESH_SHADER) != 0

        max_splats = self.positions.max_size // 12
        self.dists_buf = Buffer(
            r.device, max_splats * 4, BufferUsageFlags.STORAGE, AllocType.DEVICE, name=f"{self.name}-dists"
        )
        self.dists_alt_buf = Buffer(
            r.device, max_splats * 4, BufferUsageFlags.STORAGE, AllocType.DEVICE, name=f"{self.name}-dists-alt"
        )
        self.sorted_indices_buf = Buffer(
            r.device,
            max_splats * 4,
            BufferUsageFlags.STORAGE | BufferUsageFlags.VERTEX,
            AllocType.DEVICE,
            name=f"{self.name}-sorted-indices",
        )
        self.sorted_indices_alt_buf = Buffer(
            r.device,
            max_splats * 4,
            BufferUsageFlags.STORAGE,
            AllocType.DEVICE,
            name=f"{self.name}-sorted-indices-alt",
        )

        if not self.use_mesh_shader:
            quad_vertices = np.array([-1.0, -1.0, 1.0, -1.0, 1.0, 1.0, -1.0, 1.0], np.float32)
            quad_indices = np.array([0, 2, 1, 2, 0, 3], np.uint32)
            self.quad_vertices = Buffer.from_data(
                r.device,
                quad_vertices,
                BufferUsageFlags.VERTEX | BufferUsageFlags.TRANSFER_DST,
                AllocType.DEVICE,
                name=f"{self.name}-quad-vertices",
            )
            self.quad_indices = Buffer.from_data(
                r.device,
                quad_indices,
                BufferUsageFlags.INDEX | BufferUsageFlags.TRANSFER_DST,
                AllocType.DEVICE,
                name=f"{self.name}-quad-indices",
            )

            draw_parameters = np.array([6, max_splats, 0, 0, 0], np.uint32)
            self.draw_parameters_buf = Buffer.from_data(
                r.device,
                draw_parameters,
                BufferUsageFlags.INDIRECT | BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
                AllocType.DEVICE,
                name=f"{self.name}-draw-param",
            )
            self.draw_parameters_init = np.array([6, 0, 0, 0, 0], np.uint32).tobytes()
        else:
            mesh_workgroup_size = min(
                r.device.device_properties.mesh_shader_properties.max_preferred_mesh_work_group_invocations,
                r.device.device_properties.mesh_shader_properties.max_mesh_work_group_count[0],
            )
            draw_mesh_tasks_parameters = np.array(
                [div_round_up(max_splats, mesh_workgroup_size), 1, 1, max_splats], np.uint32
            )
            self.draw_mesh_tasks_buf = Buffer.from_data(
                r.device,
                draw_mesh_tasks_parameters,
                BufferUsageFlags.INDIRECT | BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
                AllocType.DEVICE,
                name=f"{self.name}-draw-mesh-tasks-param",
            )
            self.draw_mesh_tasks_init = np.array([0, 1, 1, 0], np.uint32).tobytes()

        defines = [
            ("DISTANCE_COMPUTE_WORKGROUP_SIZE", str(self.DISTANCE_COMPUTE_WORKGROUP_SIZE)),
            (
                "FRUSTUM_CULLING_MODE",
                str(self.FRUSTUM_CULLING_AT_DIST) if self.cull_at_dist else str(self.FRUSTUM_CULLING_AT_RASTER),
            ),
            ("USE_BARYCENTRIC", "1" if self.use_barycentric else "0"),
            (
                "MS_ANTIALIASING",
                "1" if self.mip_splatting_antialiasing else "0",
            ),  # Only useful when model comes from mip splatting
            ("MAX_SH_DEGREE", "3"),
            ("SH_FORMAT", str(self.sh_format)),
            # Using this as a flag significantly lowers performance even if disabled.
            # Likely due to worse occupancy because of higher register pressure or due to some optimization
            # not kicking in because of to the use of shader derivatives.
            ("WIREFRAME", "0"),
        ]

        if self.use_mesh_shader:
            defines.append(("RASTER_MESH_WORKGROUP_SIZE", str(mesh_workgroup_size)))

        dist_shader = r.compile_builtin_shader("3d/gaussian_splatting/dist.comp.slang", defines=defines)
        self.descriptor_layout, self.descriptor_pool, self.descriptor_sets = (
            create_descriptor_layout_pool_and_sets_ringbuffer(
                r.device,
                [DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER) for _ in range(7)],
                r.num_frames_in_flight,
            )
        )

        for s in self.descriptor_sets:
            s.write_buffer(self.dists_buf, DescriptorType.STORAGE_BUFFER, 0)
            s.write_buffer(self.sorted_indices_buf, DescriptorType.STORAGE_BUFFER, 1)
            if not self.use_mesh_shader:
                s.write_buffer(self.draw_parameters_buf, DescriptorType.STORAGE_BUFFER, 2)
            else:
                s.write_buffer(self.draw_mesh_tasks_buf, DescriptorType.STORAGE_BUFFER, 2)

        self.dist_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, dist_shader.code),
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
                self.descriptor_layout,
            ],
        )

        # Keys are UINT32 because we already convert them to being ordered when
        # interpreted as integers in the dist computation pass.
        self.sort_pipeline = GpuSortingPipeline(
            r,
            SortOptions(
                key_type=SortDataType.UINT32,
                payload_type=SortDataType.UINT32,
                indirect=self.cull_at_dist,
                unsafe_has_forward_thread_progress_guarantee=False,
            ),
        )

        if self.cull_at_dist and self.deterministic_cull_at_dist:
            self.instance_sort_pipeline = GpuSortingPipeline(
                r,
                SortOptions(
                    key_type=SortDataType.UINT32,
                    payload_type=SortDataType.UINT32,
                    indirect=True,
                    unsafe_has_forward_thread_progress_guarantee=False,
                ),
            )

        if not self.use_mesh_shader:
            vert = r.compile_builtin_shader("3d/gaussian_splatting/threedgs_raster.vert.slang", defines=defines)
            frag = r.compile_builtin_shader("3d/gaussian_splatting/threedgs_raster.frag.slang", defines=defines)
        else:
            mesh = r.compile_builtin_shader(
                "3d/gaussian_splatting/threedgs_raster.mesh.slang", target="spirv_1_4", defines=defines
            )
            frag = r.compile_builtin_shader(
                "3d/gaussian_splatting/threedgs_raster.frag.slang", entry="main_mesh", defines=defines
            )

        # Instantiate the pipeline using the compiled shaders
        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX)
                if not self.use_mesh_shader
                else PipelineStage(Shader(r.device, mesh.code), Stage.MESH),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=[
                VertexBinding(0, 8, VertexInputRate.VERTEX),
                VertexBinding(1, 4, VertexInputRate.INSTANCE),
            ]
            if not self.use_mesh_shader
            else [],
            vertex_attributes=[
                VertexAttribute(0, 0, Format.R32G32_SFLOAT),
                VertexAttribute(1, 1, Format.R32_UINT),
            ]
            if not self.use_mesh_shader
            else [],
            input_assembly=InputAssembly(
                PrimitiveTopology.TRIANGLE_LIST,
            ),
            samples=1,
            attachments=[
                Attachment(
                    format=r.output_format,
                    blend_enable=True,
                    src_color_blend_factor=BlendFactor.SRC_ALPHA,
                    dst_color_blend_factor=BlendFactor.ONE_MINUS_SRC_ALPHA,
                    color_blend_op=BlendOp.ADD,
                    src_alpha_blend_factor=BlendFactor.ONE,
                    dst_alpha_blend_factor=BlendFactor.ONE_MINUS_SRC_ALPHA,
                    alpha_blend_op=BlendOp.ADD,
                )
            ],
            depth=Depth(r.depth_format, False, False, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
                self.descriptor_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"] = mat4x3(self.current_transform_matrix)
        self.constants["inverse_transform"] = mat4x3(inverse(self.current_transform_matrix))

    def upload(self, r: Renderer, frame: RendererFrame) -> None:
        # Constants
        positions = self.positions.get_current_gpu()
        num_splats = positions.size // 12
        self.constants["splat_count"] = num_splats
        self.constants["alpha_cull_threshold"] = self.alpha_cull_threshold
        self.constants["frustum_dilation"] = self.frustum_dilation
        self.constants["splat_scale"] = self.splat_scale
        self.constants["flags"] = self.flags

        self.descriptor_set = self.descriptor_sets.get_current_and_advance()

        self.positions.get_current_gpu().write_descriptor(self.descriptor_set, DescriptorType.STORAGE_BUFFER, 3)
        self.colors.get_current_gpu().write_descriptor(self.descriptor_set, DescriptorType.STORAGE_BUFFER, 4)
        self.covariances.get_current_gpu().write_descriptor(self.descriptor_set, DescriptorType.STORAGE_BUFFER, 5)
        self.spherical_harmonics.get_current_gpu().write_descriptor(
            self.descriptor_set, DescriptorType.STORAGE_BUFFER, 6
        )

        count_buf = None
        if self.cull_at_dist:
            if self.use_mesh_shader:
                count_buf = self.draw_mesh_tasks_buf
            else:
                count_buf = self.draw_parameters_buf

            if self.deterministic_cull_at_dist:
                self.instance_sort_pipeline.upload(
                    r,
                    self.sorted_indices_buf,
                    self.sorted_indices_alt_buf,
                    self.dists_buf,
                    self.dists_alt_buf,
                    count_buf,
                )

        self.sort_pipeline.upload(
            r,
            self.dists_buf,
            self.dists_alt_buf,
            self.sorted_indices_buf,
            self.sorted_indices_alt_buf,
            count_buf,
        )

        frame.between_viewport_render_src_pipeline_stages |= PipelineStageFlags.COMPUTE_SHADER
        frame.between_viewport_render_dst_pipeline_stages |= PipelineStageFlags.TRANSFER

    def pre_render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        positions = self.positions.get_current_gpu()
        num_splats = positions.size // 12

        # Reset counters
        if self.cull_at_dist:
            if self.use_mesh_shader:
                frame.cmd.update_buffer(self.draw_mesh_tasks_buf, self.draw_mesh_tasks_init)
            else:
                frame.cmd.fill_buffer(self.draw_parameters_buf, 0, 4, 4)
            frame.upload_property_pipeline_stages |= PipelineStageFlags.COMPUTE_SHADER

        frame.cmd.memory_barrier(MemoryUsage.TRANSFER_DST, MemoryUsage.COMPUTE_SHADER)

        # Distance
        frame.cmd.bind_compute_pipeline(
            self.dist_pipeline,
            descriptor_sets=[scene_descriptor_set, self.descriptor_set],
            push_constants=self.constants.tobytes(),
        )
        frame.cmd.dispatch(div_round_up(num_splats, self.DISTANCE_COMPUTE_WORKGROUP_SIZE), 1, 1)

        frame.cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

        # Sort
        if self.cull_at_dist:
            if self.use_mesh_shader:
                count_offset = 12
            else:
                count_offset = 4

            if self.deterministic_cull_at_dist:
                self.instance_sort_pipeline.run_indirect(
                    renderer,
                    frame.cmd,
                    count_offset,
                )
                frame.cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

            self.sort_pipeline.run_indirect(
                renderer,
                frame.cmd,
                count_offset,
            )
        else:
            self.sort_pipeline.run(
                renderer,
                frame.cmd,
                num_splats,
            )

        frame.before_render_src_pipeline_stages |= PipelineStageFlags.COMPUTE_SHADER
        frame.before_render_dst_pipeline_stages |= (
            PipelineStageFlags.MESH_SHADER if self.use_mesh_shader else PipelineStageFlags.VERTEX_SHADER
        )

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=[
                self.quad_vertices,
                self.sorted_indices_buf,
            ]
            if not self.use_mesh_shader
            else [],
            index_buffer=self.quad_indices if not self.use_mesh_shader else None,
            descriptor_sets=[
                scene_descriptor_set,
                self.descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )
        if self.use_mesh_shader:
            frame.cmd.draw_mesh_tasks_indirect(self.draw_mesh_tasks_buf, 0, 1, 16)
        else:
            frame.cmd.draw_indexed_indirect(self.draw_parameters_buf, 0, 1, 20)



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

import io


def read_exact_into(file: io.FileIO, view: memoryview) -> None:
    bread = 0
    while bread < len(view):
        n = file.readinto(view[bread:])
        if n == 0:
            raise EOFError
        bread += n


def read_exact(file: io.FileIO, size: int) -> bytearray:
    out = bytearray(size)
    view = memoryview(out)
    read_exact_into(file, view)
    return out


def read_exact_at_offset_into(file: io.FileIO, offset: int, view: memoryview) -> None:
    file.seek(offset, io.SEEK_SET)
    read_exact_into(file, view)


def read_exact_at_offset(file: io.FileIO, offset: int, size: int) -> bytearray:
    file.seek(offset, io.SEEK_SET)
    return read_exact(file, size)



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from contextlib import contextmanager
from time import perf_counter_ns


@contextmanager
def profile(name: str):  # type: ignore
    begin = perf_counter_ns()
    yield
    end = perf_counter_ns()
    delta = end - begin
    print(f"{name}: {delta * 1e-6:.3f}ms")  # noqa: T201



import logging
import struct
import sys

import numpy as np
from pyglm.glm import ivec2, normalize, quatLookAtRH, vec3
from pyxpg import *

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig, RendererConfig, UploadMethod
from ambra.lights import DirectionalLight, DirectionalShadowSettings
from ambra.primitives3d import Mesh
from ambra.property import BufferProperty, UploadSettings
from ambra.utils.descriptors import create_descriptor_layout_pool_and_set
from ambra.utils.hook import hook
from ambra.utils.io import (
    read_exact,
    read_exact_at_offset,
    read_exact_at_offset_into,
)
from ambra.viewer import Action, Key, Modifiers, Viewer, imgui

logging.basicConfig(
    level=logging.NOTSET,
    format="[%(asctime)s.%(msecs)03d] %(levelname)-6s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

NUM_WORKERS = 4

files = [open(sys.argv[1], "rb", buffering=0) for _ in range(NUM_WORKERS)]
file = files[0]
header = read_exact(file, 12)
N = struct.unpack("<I", header[0:4])[0]
V = struct.unpack("<I", header[4:8])[0]
I = struct.unpack("<I", header[8:12])[0]

indices = np.frombuffer(read_exact_at_offset(file, N * V * 12 + len(header), I * 4), np.uint32)

scale = 1.0


class FileStreamingProperty(BufferProperty):
    def _get_size_offset(frame_index: int):
        return V * 12, 12 + frame_index * V * 12

    def get_frame_by_index_into(self, frame_index: int, out: memoryview, thread_index: int = -1) -> int:
        size, offset = FileStreamingProperty._get_size_offset(frame_index)
        read_exact_at_offset_into(files[thread_index], offset, out[:size])
        return size

    # def get_frame_by_index(self, frame_index: int, thread_index: int = -1):
    #     size, offset = FileStreamingProperty._get_size_offset(frame_index)
    #     buf = np.empty(size, np.uint8)
    #     print(offset, frame_index)
    #     read_exact_at_offset_into(files[thread_index], offset, buf.data)
    #     return buf.view(np.float32).reshape((-1, 3)) * scale


positions = FileStreamingProperty(
    V * 12,
    N,
    np.float32,
    (-1, 3),
    upload=UploadSettings(
        preupload=False,
        async_load=True,
        cpu_prefetch_count=2,
        gpu_prefetch_count=2,
    ),
)

mesh = Mesh(positions, indices=indices)


class CustomViewer(Viewer):
    def __init__(self, title="ambra", config=None, key_map=None):
        super().__init__(title, config, key_map)
        self._texture = None

    @hook
    def on_gui(self):
        global scale
        if imgui.begin("Control"):
            u, scale = imgui.slider_float("SCALE", scale, 0.1, 10.0)
            if u:
                mesh.positions.update_frame(self.playback.current_frame, None)
        imgui.end()
        if imgui.begin("Image"):
            if hasattr(light, "shadow_map"):
                if self._texture is None and light.shadow_map is not None:
                    sampler = Sampler(
                        viewer.device,
                        u=SamplerAddressMode.REPEAT,
                        v=SamplerAddressMode.REPEAT,
                    )
                    layout, pool, set = create_descriptor_layout_pool_and_set(
                        viewer.device,
                        [
                            DescriptorSetBinding(1, DescriptorType.COMBINED_IMAGE_SAMPLER, stage_flags=Stage.FRAGMENT),
                        ],
                    )
                    set.write_combined_image_sampler(
                        light.shadow_map, ImageLayout.SHADER_READ_ONLY_OPTIMAL, sampler, 0
                    )
                    self._sampler = sampler
                    self._texture = imgui.Texture(set)

                if self._texture is not None:
                    avail = imgui.get_content_region_avail()
                    available = ivec2(avail.x, avail.y)

                    ar = 1.0

                    # height = available.x / ar
                    height = available.y
                    view_size = ivec2(ar * height, height)
                    imgui.image(
                        self._texture,
                        imgui.Vec2(*view_size),
                        uv1=(1, -1),
                    )
        imgui.end()

    def on_key(self, key: Key, action: Action, modifiers: Modifiers):
        handled = False
        if action == Action.PRESS or action == Action.REPEAT:
            if key == Key.A:
                mesh.positions.update_frame(self.playback.current_frame, None)
        if not handled:
            return super().on_key(key, action, modifiers)


viewer = CustomViewer(
    config=Config(
        # preferred_frames_in_flight=3,
        playback=PlaybackConfig(
            playing=True,
            frames_per_second=25.0,
        ),
        renderer=RendererConfig(
            background_color=(0, 0, 0, 1),
            # force_buffer_upload_method=UploadMethod.GRAPHICS_QUEUE,
            # force_buffer_upload_method=UploadMethod.TRANSFER_QUEUE,
            # force_buffer_upload_method=UploadMethod.MAPPED_PREFER_HOST,
            # force_buffer_upload_method=UploadMethod.MAPPED_PREFER_DEVICE,
            upload_buffer_count=2,
            thread_pool_workers=NUM_WORKERS,
        ),
        gui=GuiConfig(
            stats=True,
            playback=True,
            inspector=True,
            renderer=True,
        ),
        camera=CameraConfig(
            position=(10, -10, 10),
            target=(0, 0, 0),
        ),
        world_up=(0, -1, 0),
    ),
)

light = DirectionalLight.look_at(
    vec3(5, -5, 5),
    vec3(0, 0, 0),
    vec3(0, 1, 0),
    np.array([1.0, 1.0, 1.0]),
    shadow_settings=DirectionalShadowSettings(half_extent=5.0),
)

viewer.scene.objects.append(mesh)
viewer.scene.objects.append(light)

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from typing import Any, Callable


class hook:  # noqa: N801
    def __init__(self, fn: Callable[[Any], Any]) -> None:
        self.fn = fn

    def __set_name__(self, owner: Any, name: str) -> None:
        func = self.fn

        def _decorator(self: Any, *args: Any, **kwargs: Any) -> Any:
            super_obj = super(owner, self)
            super_fn = getattr(super_obj, func.__name__)
            super_fn(*args, **kwargs)
            return func(self, *args, **kwargs)

        setattr(owner, name, _decorator)

    def __call__(self) -> None:
        raise AssertionError(
            "@hook_after decorator object should never be called directly. This can happen if you apply this decorator to a function that is not a method."
        )


class hook_before:  # noqa: N801
    def __init__(self, fn: Callable[[Any], Any]) -> None:
        self.fn = fn

    def __set_name__(self, owner: Any, name: str) -> None:
        func = self.fn

        def _decorator(self: Any, *args: Any, **kwargs: Any) -> Any:
            super_obj = super(owner, self)
            super_fn = getattr(super_obj, func.__name__)
            result = func(self, *args, **kwargs)
            super_fn(*args, **kwargs)
            return result

        setattr(owner, name, _decorator)

    def __call__(self) -> None:
        raise AssertionError(
            "@hook decorator object should never be called directly. This can happen if you apply this decorator to a function that is not a method."
        )



import numpy as np
from pyglm.glm import normalize, vec4
from pyxpg import imgui

from ambra.config import Config, GuiConfig, PlaybackConfig
from ambra.primitives3d import Grid, GridType
from ambra.viewer import Viewer


class CustomViewer(Viewer):
    def on_gui(self):
        global grid
        if imgui.begin("Hello")[0]:
            u, t = imgui.drag_float3("Translation", grid.translation.get_current(), 0.01)
            if u:
                grid.translation.update_frame(0, np.array(t))

            u, s = imgui.drag_float3("Scale", grid.scale.get_current(), 0.01)
            if u:
                grid.scale.update_frame(0, np.array(s))

            u, r = imgui.drag_float4("Rotation", grid.rotation.get_current(), 0.1)
            if u:
                grid.rotation.update_frame(0, np.array(normalize(vec4(r))))
        imgui.end()

        return super().on_gui()


viewer = CustomViewer(
    config=Config(
        gui=GuiConfig(
            stats=True,
        ),
    ),
)

grid = Grid.transparent_black_lines((100, 100), GridType.XY_PLANE)

viewer.scene.objects.extend([grid])
viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, List, Optional

import numpy as np
from numpy.typing import NDArray
from pyglm.glm import inverse, normalize, orthoRH_ZO, quatLookAtRH, vec3, vec4
from pyxpg import (
    AccessFlags,
    AllocType,
    Buffer,
    BufferUsageFlags,
    ComputePipeline,
    DepthAttachment,
    DescriptorSetBinding,
    DescriptorType,
    Device,
    Filter,
    Format,
    Image,
    ImageAspectFlags,
    ImageBarrier,
    ImageLayout,
    ImageUsageFlags,
    ImageView,
    ImageViewType,
    LoadOp,
    MemoryUsage,
    PipelineStageFlags,
    PushConstantsRange,
    Shader,
    StoreOp,
)

from .property import BufferProperty
from .renderer_frame import RendererFrame
from .scene import Object, Object3D
from .utils.descriptors import (
    create_descriptor_layout_pool_and_set,
    create_descriptor_layout_pool_and_sets,
    create_descriptor_pool_and_sets_ringbuffer,
)
from .utils.gpu import UploadableBuffer, view_bytes
from .utils.ring_buffer import RingBuffer

if TYPE_CHECKING:
    from .renderer import Renderer


class LightTypes(Enum):
    DIRECTIONAL = 0


@dataclass
class LightInfo:
    size: int


directional_light_dtype = np.dtype(
    {
        "orthographic_camera": (np.dtype((np.float32, (4, 4))), 0),
        "radiance": (np.dtype((np.float32, (3,))), 64),
        "shadow_map_index": (np.int32, 76),
        "direction": (np.dtype((np.float32, (3,))), 80),
        "max_bias": (np.float32, 92),
        "min_bias": (np.float32, 96),
        "_padding": (np.dtype((np.float32, (3,))), 100),
    }
)  # type: ignore

# When adding a new light type, this also has to be added with a matching type to "shaders/2d/scene.slang" and "shaders/3d/scene.slang"
LIGHT_TYPES_INFO = [
    LightInfo(directional_light_dtype.itemsize),
]


class Light(Object3D):
    def render_shadow_maps(self, renderer: "Renderer", frame: RendererFrame, objects: List[Object]) -> None:
        pass


class PointLight(Light):
    def __init__(
        self,
        intensity: BufferProperty,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        super().__init__(name, translation, rotation, scale, enabled=enabled)
        self.intensity = self.add_buffer_property(intensity, np.float32, (-1, 3), name="intensity")


class SpotLight(Light):
    def __init__(
        self,
        intensity: BufferProperty,
        stop_cosine: BufferProperty,
        falloff_start_cosine: BufferProperty,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        super().__init__(name, translation, rotation, scale, enabled=enabled)
        self.intensity = self.add_buffer_property(intensity, np.float32, (-1, 3), name="intensity")
        self.stop_cosine = self.add_buffer_property(stop_cosine, np.float32, (-1, 3), name="stop_cosine")
        self.falloff_start_cosine = self.add_buffer_property(
            falloff_start_cosine, np.float32, (-1, 3), name="falloff_start_cosine"
        )


@dataclass
class DirectionalShadowSettings:
    casts_shadow: bool = True
    shadow_map_size: int = 2048
    half_extent: float = 100.0
    z_near: float = 0.0
    z_far: float = 1000.0
    max_bias: float = 0.003
    min_bias: float = 0.001


class DirectionalLight(Light):
    def __init__(
        self,
        radiance: BufferProperty,
        shadow_settings: Optional[DirectionalShadowSettings] = None,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)
        self.radiance = self.add_buffer_property(radiance, np.float32, (3,), name="radiance")
        self.shadow_settings = shadow_settings if shadow_settings is not None else DirectionalShadowSettings()
        self.shadow_map: Optional[Image] = None

    @classmethod
    def look_at(
        cls,
        position: vec3,
        target: vec3,
        world_up: vec3,
        radiance: BufferProperty,
        shadow_settings: Optional[DirectionalShadowSettings] = None,
        name: Optional[str] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ) -> "DirectionalLight":
        rotation = quatLookAtRH(normalize(target - position), world_up)
        return cls(
            radiance,
            shadow_settings,
            name,
            translation=np.asarray(position),  # type: ignore
            rotation=np.asarray(rotation),  # type: ignore
            scale=scale,
            enabled=enabled,
            viewport_mask=viewport_mask,
        )

    def create(self, r: "Renderer") -> None:
        if self.shadow_settings.casts_shadow:
            self.shadow_map = Image(
                r.device,
                self.shadow_settings.shadow_map_size,
                self.shadow_settings.shadow_map_size,
                r.shadow_map_format,
                ImageUsageFlags.SAMPLED | ImageUsageFlags.DEPTH_STENCIL_ATTACHMENT,
                AllocType.DEVICE,
                name=f"{self.name}-shadow_map",
            )
            self.shadow_map_viewport = [
                0,
                0,
                self.shadow_settings.shadow_map_size,
                self.shadow_settings.shadow_map_size,
            ]

            self.descriptor_pool, self.descriptor_sets = create_descriptor_pool_and_sets_ringbuffer(
                r.device, r.scene_depth_descriptor_set_layout, r.num_frames_in_flight, name="scene-descriptors"
            )

            constants_dtype = np.dtype(
                {
                    "camera_matrix": (np.dtype((np.float32, (4, 4))), 0),
                }
            )  # type: ignore

            self.constants = np.zeros((1,), constants_dtype)

            self.uniform_buffers = RingBuffer(
                [
                    UploadableBuffer(r.device, constants_dtype.itemsize, BufferUsageFlags.UNIFORM)
                    for _ in range(r.num_frames_in_flight)
                ]
            )
            for set, buf in zip(self.descriptor_sets, self.uniform_buffers):
                set.write_buffer(buf, DescriptorType.UNIFORM_BUFFER, 0, 0)

        # TODO: this and other matrices should be using config to know what is the deafult data handedness
        # we should also have default front/back face winding andr require dynamic state for culling mode.
        self.projection = orthoRH_ZO(
            -self.shadow_settings.half_extent,
            self.shadow_settings.half_extent,
            self.shadow_settings.half_extent,
            -self.shadow_settings.half_extent,
            self.shadow_settings.z_near,
            self.shadow_settings.z_far,
        )

        self.light_info = np.zeros((1,), directional_light_dtype)

    def upload_light(
        self, renderer: "Renderer", frame: RendererFrame, shadow_map_index: int, offset: int, buffer: UploadableBuffer
    ) -> None:
        view = inverse(self.current_transform_matrix)
        direction = vec3(self.current_transform_matrix * vec4(0, 0, -1, 0))
        self.light_info["orthographic_camera"] = self.projection * view
        self.light_info["radiance"] = self.radiance.get_current()
        self.light_info["shadow_map_index"] = shadow_map_index
        self.light_info["direction"] = normalize(direction)
        self.light_info["max_bias"] = self.shadow_settings.max_bias
        self.light_info["min_bias"] = self.shadow_settings.min_bias

        buffer.upload(frame.cmd, MemoryUsage.NONE, view_bytes(self.light_info), offset)

    def upload(self, renderer: "Renderer", frame: RendererFrame) -> None:
        if not self.shadow_settings.casts_shadow or frame.path_tracing:
            return

        assert self.shadow_map is not None
        view = inverse(self.current_transform_matrix)
        self.constants["camera_matrix"] = self.projection * view
        buf = self.uniform_buffers.get_current_and_advance()
        buf.upload(
            frame.cmd,
            MemoryUsage.NONE,  # Synchronized by automatic after-upload barrier
            self.constants.view(np.uint8),
        )
        frame.upload_property_pipeline_stages |= PipelineStageFlags.VERTEX_SHADER

        frame.upload_after_image_barriers.append(
            ImageBarrier(
                self.shadow_map,
                ImageLayout.UNDEFINED,
                ImageLayout.DEPTH_STENCIL_ATTACHMENT_OPTIMAL,
                PipelineStageFlags.FRAGMENT_SHADER,
                AccessFlags.SHADER_SAMPLED_READ,
                PipelineStageFlags.EARLY_FRAGMENT_TESTS,
                AccessFlags.DEPTH_STENCIL_ATTACHMENT_READ | AccessFlags.DEPTH_STENCIL_ATTACHMENT_WRITE,
                aspect_mask=ImageAspectFlags.DEPTH,
            )
        )

        frame.before_render_image_barriers.append(
            ImageBarrier(
                self.shadow_map,
                ImageLayout.DEPTH_STENCIL_ATTACHMENT_OPTIMAL,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                PipelineStageFlags.LATE_FRAGMENT_TESTS,
                AccessFlags.DEPTH_STENCIL_ATTACHMENT_READ | AccessFlags.DEPTH_STENCIL_ATTACHMENT_WRITE,
                PipelineStageFlags.FRAGMENT_SHADER,
                AccessFlags.SHADER_SAMPLED_READ,
                aspect_mask=ImageAspectFlags.DEPTH,
            )
        )

    def render_shadow_maps(self, renderer: "Renderer", frame: RendererFrame, objects: List[Object]) -> None:
        if not self.shadow_settings.casts_shadow:
            return

        assert self.shadow_map is not None

        descriptor_set = self.descriptor_sets.get_current_and_advance()

        frame.cmd.set_viewport(self.shadow_map_viewport)
        frame.cmd.set_scissors(self.shadow_map_viewport)
        with frame.cmd.rendering(
            self.shadow_map_viewport, [], DepthAttachment(self.shadow_map, LoadOp.CLEAR, StoreOp.STORE, 1.0)
        ):
            for o in objects:
                o.render_depth(renderer, frame, descriptor_set)


class UniformEnvironmentLight(Light):
    def __init__(
        self,
        radiance: BufferProperty,
        name: Optional[str] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        super().__init__(name, enabled=enabled, viewport_mask=viewport_mask)
        self.radiance = self.add_buffer_property(radiance, np.float32, (3,), name="radiance")


@dataclass(frozen=True)
class IBLParams:
    # Skybox
    skybox_size: int = 1024

    # Irradiance
    irradiance_size: int = 32
    irradiance_samples_phi: int = 256
    irradiance_samples_theta: int = 64

    # Specular
    specular_size: int = 1024
    specular_mips: int = 5
    specular_samples: int = 1024


@dataclass
class EnvironmentCubemaps:
    irradiance_cubemap: NDArray[np.float16]
    specular_cubemap: List[NDArray[np.float16]]

    @classmethod
    def load(cls, file: Any, **kwargs: Any) -> "EnvironmentCubemaps":
        arrays = np.load(file, allow_pickle=False, **kwargs)
        irradiance_cubemap = arrays["irradiance"]
        specular_cubemap = []
        for i in range(16):
            k = f"specular_{i}"
            if k not in arrays:
                break
            specular_cubemap.append(arrays[k])
        if len(specular_cubemap) == 0:
            raise KeyError('no specular cubemap ("specular_0") found in the archive')
        return cls(irradiance_cubemap, specular_cubemap)

    def save(self, file: Any, **kwargs: Any) -> None:
        np.savez(
            file,
            allow_pickle=False,
            irradiance=self.irradiance_cubemap,
            **{f"specular_{i}": s for i, s in enumerate(self.specular_cubemap)},
            **kwargs,
        )

    def gpu(self, device: Device) -> "GpuEnvironmentCubemaps":
        # Create cubemaps
        irradiance_size = self.irradiance_cubemap.shape[1]
        irradiance_cubemap = Image(
            device,
            irradiance_size,
            irradiance_size,
            Format.R16G16B16A16_SFLOAT,
            ImageUsageFlags.SAMPLED
            | ImageUsageFlags.STORAGE
            | ImageUsageFlags.TRANSFER_SRC
            | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            array_layers=6,
            is_cube=True,
        )
        specular_cubemap = Image(
            device,
            self.specular_cubemap[0].shape[2],
            self.specular_cubemap[0].shape[1],
            Format.R16G16B16A16_SFLOAT,
            ImageUsageFlags.SAMPLED
            | ImageUsageFlags.STORAGE
            | ImageUsageFlags.TRANSFER_SRC
            | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            array_layers=6,
            mip_levels=len(self.specular_cubemap),
            is_cube=True,
        )

        # Create upload buffers
        irradiance_buffer = Buffer.from_data(
            device, self.irradiance_cubemap, BufferUsageFlags.TRANSFER_SRC, AllocType.HOST
        )
        specular_buffers = [
            Buffer.from_data(device, a, BufferUsageFlags.TRANSFER_SRC, AllocType.HOST) for a in self.specular_cubemap
        ]

        # Upload
        with device.sync_commands() as cmd:
            cmd.image_barrier(
                irradiance_cubemap, ImageLayout.TRANSFER_DST_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_DST
            )
            cmd.copy_buffer_to_image_range(
                irradiance_buffer, irradiance_cubemap, irradiance_size, irradiance_size, image_layer_count=6
            )
            cmd.image_barrier(
                irradiance_cubemap,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                MemoryUsage.TRANSFER_DST,
                MemoryUsage.SHADER_READ_ONLY,
            )

            cmd.image_barrier(
                specular_cubemap, ImageLayout.TRANSFER_DST_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_DST
            )
            for m, (b, a) in enumerate(zip(specular_buffers, self.specular_cubemap)):
                cmd.copy_buffer_to_image_range(
                    b, specular_cubemap, a.shape[1], a.shape[1], image_mip=m, image_layer_count=6
                )
            cmd.image_barrier(
                specular_cubemap,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                MemoryUsage.TRANSFER_DST,
                MemoryUsage.SHADER_READ_ONLY,
            )

        return GpuEnvironmentCubemaps(irradiance_cubemap, specular_cubemap)


@dataclass
class GpuEnvironmentCubemaps:
    irradiance_cubemap: Image
    specular_cubemap: Image

    def cpu(self, device: Device) -> EnvironmentCubemaps:
        # Create buffers
        irradiance_size = self.irradiance_cubemap.width
        irradiance_shape = (6, irradiance_size, irradiance_size, 4)
        irradiance_buffer = Buffer(
            device,
            np.prod(irradiance_shape) * 2,  # type: ignore
            BufferUsageFlags.TRANSFER_DST,
            AllocType.HOST,
        )

        specular_size = self.specular_cubemap.width
        specular_mips = self.specular_cubemap.mip_levels
        specular_buffers = []
        for m in range(specular_mips):
            specular_mip_size = specular_size >> m
            specular_mip_shape = (6, specular_mip_size, specular_mip_size, 4)
            specular_buffers.append(
                Buffer(device, np.prod(specular_mip_shape) * 2, BufferUsageFlags.TRANSFER_DST, AllocType.HOST)  # type: ignore
            )

        # Readback
        with device.sync_commands() as cmd:
            cmd.image_barrier(
                self.irradiance_cubemap, ImageLayout.TRANSFER_SRC_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_SRC
            )
            cmd.copy_image_to_buffer_range(
                self.irradiance_cubemap, irradiance_buffer, irradiance_size, irradiance_size, image_layer_count=6
            )
            cmd.image_barrier(
                self.irradiance_cubemap,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                MemoryUsage.TRANSFER_SRC,
                MemoryUsage.SHADER_READ_ONLY,
            )

            cmd.image_barrier(
                self.specular_cubemap, ImageLayout.TRANSFER_SRC_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_SRC
            )
            for m, b in enumerate(specular_buffers):
                specular_mip_size = specular_size >> m
                cmd.copy_image_to_buffer_range(
                    self.specular_cubemap, b, specular_mip_size, specular_mip_size, image_mip=m, image_layer_count=6
                )
            cmd.image_barrier(
                self.specular_cubemap,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                MemoryUsage.TRANSFER_SRC,
                MemoryUsage.SHADER_READ_ONLY,
            )
            cmd.memory_barrier(MemoryUsage.TRANSFER_SRC, MemoryUsage.HOST_READ)

        # Copy into numpy arrays
        irradiance_array = np.frombuffer(irradiance_buffer.data, np.float16).copy().reshape(irradiance_shape)
        specular_arrays: List[NDArray[np.float16]] = []
        for m, b in enumerate(specular_buffers):
            specular_mip_size = specular_size >> m
            specular_mip_shape = (6, specular_mip_size, specular_mip_size, 4)
            specular_arrays.append(np.frombuffer(b, np.float16).copy().reshape(specular_mip_shape))

        return EnvironmentCubemaps(irradiance_array, specular_arrays)


class GGXLUTPipeline:
    def __init__(self, r: "Renderer"):
        self.lut_constants_dtype = np.dtype(
            {
                "samples": (np.uint32, 0),
            }
        )  # type: ignore
        self.lut_constants = np.zeros((1,), self.lut_constants_dtype)
        self.lut_shader = r.compile_builtin_shader("3d/ibl.slang", "entry_lut")
        self.descriptor_set_layout, self.descriptor_pool, self.descriptor_set = create_descriptor_layout_pool_and_set(
            r.device,
            [
                DescriptorSetBinding(1, DescriptorType.STORAGE_IMAGE),
            ],
        )
        self.lut_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.lut_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.lut_constants_dtype.itemsize)],
        )

    def run(self, r: "Renderer", width: int = 512, height: int = 512, samples: int = 1024) -> Image:
        lut = Image(
            r.device,
            width,
            height,
            Format.R16G16_SFLOAT,
            ImageUsageFlags.SAMPLED
            | ImageUsageFlags.STORAGE
            | ImageUsageFlags.TRANSFER_SRC
            | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
        )

        self.lut_constants["samples"] = samples

        descriptor_set = self.descriptor_set
        descriptor_set.write_image(lut, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0)

        with r.device.sync_commands() as cmd:
            cmd.image_barrier(lut, ImageLayout.GENERAL, MemoryUsage.NONE, MemoryUsage.COMPUTE_SHADER, undefined=True)
            cmd.bind_compute_pipeline(
                self.lut_pipeline, descriptor_sets=[descriptor_set], push_constants=self.lut_constants.tobytes()
            )
            cmd.dispatch((width + 7) // 8, (height + 7) // 8, 6)
            cmd.image_barrier(lut, ImageLayout.SHADER_READ_ONLY_OPTIMAL, MemoryUsage.COMPUTE_SHADER, MemoryUsage.ALL)

        return lut


class IBLPipeline:
    def __init__(self, r: "Renderer"):
        # Common
        self.max_specular_mips = 16
        self.descriptor_set_layout, self.descriptor_pool, self.descriptor_sets = (
            create_descriptor_layout_pool_and_sets(
                r.device,
                [
                    DescriptorSetBinding(1, DescriptorType.STORAGE_IMAGE),
                    DescriptorSetBinding(1, DescriptorType.COMBINED_IMAGE_SAMPLER),
                ],
                self.max_specular_mips,
            )
        )

        # Skybox
        self.skybox_constants_dtype = np.dtype(
            {
                "size": (np.uint32, 0),
            }
        )  # type: ignore
        self.skybox_constants = np.zeros((1,), self.skybox_constants_dtype)
        self.skybox_shader = r.compile_builtin_shader("3d/ibl.slang", "entry_skybox")
        self.skybox_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.skybox_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.skybox_constants.itemsize)],
        )

        # Irradiance
        self.irradiance_constants_dtype = np.dtype(
            {
                "size": (np.uint32, 0),
                "samples_phi": (np.uint32, 4),
                "samples_theta": (np.uint32, 8),
            }
        )  # type: ignore
        self.irradiance_constants = np.zeros((1,), self.irradiance_constants_dtype)
        self.irradiance_shader = r.compile_builtin_shader("3d/ibl.slang", "entry_irradiance")
        self.irradiance_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.irradiance_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.irradiance_constants.itemsize)],
        )

        # Specular
        self.specular_constants_dtype = np.dtype(
            {
                "size": (np.uint32, 0),
                "samples": (np.uint32, 4),
                "roughness": (np.float32, 8),
                "skybox_resolution": (np.float32, 12),
            }
        )  # type: ignore
        self.specular_constants = np.zeros((1,), self.specular_constants_dtype)
        self.specular_shader = r.compile_builtin_shader("3d/ibl.slang", "entry_specular")
        self.specular_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.specular_shader.code),
            descriptor_set_layouts=[self.descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.specular_constants_dtype.itemsize)],
        )

    def run(self, r: "Renderer", equirectangular: Image, params: IBLParams) -> GpuEnvironmentCubemaps:
        skybox_size = params.skybox_size
        skybox_mip_levels = skybox_size.bit_length()
        skybox_cubemap = Image(
            r.device,
            skybox_size,
            skybox_size,
            Format.R16G16B16A16_SFLOAT,
            ImageUsageFlags.SAMPLED
            | ImageUsageFlags.STORAGE
            | ImageUsageFlags.TRANSFER_SRC
            | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            array_layers=6,
            mip_levels=skybox_mip_levels,
            is_cube=True,
        )

        # Skybox
        descriptor_set = self.descriptor_sets[0]
        self.skybox_constants["size"] = params.skybox_size
        cubemap_array_view = ImageView(r.device, skybox_cubemap, ImageViewType.TYPE_2D_ARRAY, mip_level_count=1)
        descriptor_set.write_image(cubemap_array_view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0)
        descriptor_set.write_combined_image_sampler(
            equirectangular, ImageLayout.SHADER_READ_ONLY_OPTIMAL, r.linear_sampler, 1
        )

        with r.device.sync_commands() as cmd:
            cmd.image_barrier(
                skybox_cubemap, ImageLayout.GENERAL, MemoryUsage.NONE, MemoryUsage.COMPUTE_SHADER, undefined=True
            )
            cmd.bind_compute_pipeline(
                self.skybox_pipeline, descriptor_sets=[descriptor_set], push_constants=self.skybox_constants.tobytes()
            )
            cmd.dispatch((skybox_size + 7) // 8, (skybox_size + 7) // 8, 6)
            cmd.image_barrier(skybox_cubemap, ImageLayout.GENERAL, MemoryUsage.COMPUTE_SHADER, MemoryUsage.ALL)
            for m in range(1, skybox_mip_levels):
                src_size = skybox_size >> m - 1
                dst_size = skybox_size >> m
                cmd.blit_image_range(
                    skybox_cubemap,
                    skybox_cubemap,
                    src_size,
                    src_size,
                    dst_size,
                    dst_size,
                    Filter.LINEAR,
                    src_mip_level=m - 1,
                    dst_mip_level=m,
                    src_array_layer_count=6,
                    dst_array_layer_count=6,
                )
                cmd.memory_barrier(MemoryUsage.ALL, MemoryUsage.ALL)
            cmd.image_barrier(
                skybox_cubemap, ImageLayout.SHADER_READ_ONLY_OPTIMAL, MemoryUsage.COMPUTE_SHADER, MemoryUsage.ALL
            )

        # Irradiance
        irradiance_size = params.irradiance_size
        irradiance_cubemap = Image(
            r.device,
            irradiance_size,
            irradiance_size,
            Format.R16G16B16A16_SFLOAT,
            ImageUsageFlags.SAMPLED | ImageUsageFlags.STORAGE | ImageUsageFlags.TRANSFER_SRC,
            AllocType.DEVICE,
            array_layers=6,
            is_cube=True,
        )

        self.irradiance_constants["size"] = irradiance_size
        self.irradiance_constants["samples_phi"] = params.irradiance_samples_phi
        self.irradiance_constants["samples_theta"] = params.irradiance_samples_theta

        cubemap_array_view = ImageView(r.device, irradiance_cubemap, ImageViewType.TYPE_2D_ARRAY, mip_level_count=1)
        descriptor_set.write_image(cubemap_array_view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0)
        descriptor_set.write_combined_image_sampler(
            skybox_cubemap, ImageLayout.SHADER_READ_ONLY_OPTIMAL, r.linear_sampler, 1
        )

        with r.device.sync_commands() as cmd:
            cmd.image_barrier(
                irradiance_cubemap, ImageLayout.GENERAL, MemoryUsage.NONE, MemoryUsage.COMPUTE_SHADER, undefined=True
            )
            cmd.bind_compute_pipeline(
                self.irradiance_pipeline,
                descriptor_sets=[descriptor_set],
                push_constants=self.irradiance_constants.tobytes(),
            )
            cmd.dispatch((irradiance_size + 7) // 8, (irradiance_size + 7) // 8, 6)
            cmd.image_barrier(
                irradiance_cubemap,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                MemoryUsage.COMPUTE_SHADER,
                MemoryUsage.ALL,
                undefined=False,
            )

        # Specular
        specular_mips = min(self.max_specular_mips, params.specular_mips, params.specular_size.bit_length())
        specular_size = params.specular_size
        specular_cubemap = Image(
            r.device,
            params.specular_size,
            params.specular_size,
            Format.R16G16B16A16_SFLOAT,
            ImageUsageFlags.SAMPLED | ImageUsageFlags.STORAGE | ImageUsageFlags.TRANSFER_SRC,
            AllocType.DEVICE,
            array_layers=6,
            mip_levels=specular_mips,
            is_cube=True,
        )

        self.specular_constants["samples"] = params.specular_samples
        self.specular_constants["skybox_resolution"] = skybox_size
        views = []
        with r.device.sync_commands() as cmd:
            cmd.image_barrier(
                specular_cubemap, ImageLayout.GENERAL, MemoryUsage.NONE, MemoryUsage.COMPUTE_SHADER, undefined=True
            )

            for m in range(specular_mips):
                descriptor_set = self.descriptor_sets[m]
                specular_mip_view = ImageView(
                    r.device, specular_cubemap, ImageViewType.TYPE_2D_ARRAY, base_mip_level=m, mip_level_count=1
                )
                descriptor_set.write_image(specular_mip_view, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0)
                descriptor_set.write_combined_image_sampler(
                    skybox_cubemap, ImageLayout.SHADER_READ_ONLY_OPTIMAL, r.linear_sampler, 1
                )
                views.append(specular_mip_view)

                mip_size = specular_size >> m
                self.specular_constants["size"] = mip_size
                self.specular_constants["roughness"] = m / (specular_mips - 1)
                cmd.bind_compute_pipeline(
                    self.specular_pipeline,
                    descriptor_sets=[descriptor_set],
                    push_constants=self.specular_constants.tobytes(),
                )
                cmd.dispatch((mip_size + 7) // 8, (mip_size + 7) // 8, 6)

            cmd.image_barrier(specular_cubemap, ImageLayout.SHADER_READ_ONLY_OPTIMAL, MemoryUsage.ALL, MemoryUsage.ALL)

        return GpuEnvironmentCubemaps(irradiance_cubemap, specular_cubemap)


class EnvironmentLight(Light):
    def __init__(
        self,
        equirectangular: Optional[NDArray[np.float32]] = None,
        cubemaps: Optional[EnvironmentCubemaps] = None,
        ibl_params: Optional[IBLParams] = None,
        name: Optional[str] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        if not ((equirectangular is None) ^ (cubemaps is None)):
            raise RuntimeError('Exactly one of "equirectangular" and "cubemaps" must not be None')

        if equirectangular is not None:
            height, width, channels = equirectangular.shape
            if not channels == 3 or not (equirectangular.dtype == np.float32 or equirectangular.dtype == np.float64):
                raise ValueError(
                    f"Equirectangular map must have 3 channels and be of float32 or float64 dtype. Got {channels} channels and {equirectangular.dtype} dtype."
                )
            # TODO: check if RGB_F32 is supported instead of defaulting to RGBA_F32
            rgb_data = equirectangular.astype(np.float32, copy=False)
            equirectangular = np.dstack((rgb_data, np.ones((height, width, 1), np.float32)))

        self.equirectangular = equirectangular
        self.cubemaps = cubemaps
        self.ibl_params = ibl_params
        super().__init__(name, enabled=enabled, viewport_mask=viewport_mask)

    def create(self, r: "Renderer") -> None:
        if self.equirectangular is not None:
            height, width, _channels = self.equirectangular.shape
            equirectangular_img = Image.from_data(
                r.device,
                self.equirectangular,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                width,
                height,
                Format.R32G32B32A32_SFLOAT,
                ImageUsageFlags.SAMPLED | ImageUsageFlags.TRANSFER_DST,
                AllocType.DEVICE,
            )
            self.gpu_cubemaps = r.run_ibl_pipeline(equirectangular_img, self.ibl_params)
        else:
            assert self.cubemaps is not None
            self.gpu_cubemaps = self.cubemaps.gpu(r.device)

    @classmethod
    def from_equirectangular(
        cls,
        equirectangular: NDArray[np.float32],
        ibl_params: Optional[IBLParams] = None,
        name: Optional[str] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ) -> "EnvironmentLight":
        return cls(
            equirectangular=equirectangular,
            ibl_params=ibl_params,
            name=name,
            enabled=enabled,
            viewport_mask=viewport_mask,
        )

    @classmethod
    def from_cubemaps(
        cls,
        cubemaps: EnvironmentCubemaps,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ) -> "EnvironmentLight":
        return cls(cubemaps=cubemaps, enabled=enabled, viewport_mask=viewport_mask)


# class AreaLight(Light):
#     pass



import sys

import numpy as np
import PIL.Image
from pyxpg import *

from ambra.config import Config
from ambra.utils.gpu import MipGenerationFilter, readback_image_mips
from ambra.viewer import Viewer

if False:
    data = np.array(
        [
            [
                [127, 127, 127, 255],
                [127, 127, 127, 255],
                [127, 127, 127, 255],
                [127, 127, 127, 255],
            ],
            [
                [127, 127, 127, 255],
                [127, 127, 127, 255],
                [127, 127, 127, 255],
                [127, 127, 127, 255],
            ],
        ],
        np.uint8,
    )
elif False:
    data = np.array(
        [
            [
                [255, 255, 255, 255],
                [0, 0, 0, 255],
            ],
            [
                [0, 0, 0, 255],
                [255, 255, 255, 255],
            ],
        ],
        np.uint8,
    )
else:
    data = np.array(
        [
            [
                [255, 255, 255, 255],
                [255, 255, 255, 255],
                [0, 0, 0, 255],
                [0, 0, 0, 255],
            ],
            [
                [255, 255, 255, 255],
                [255, 255, 255, 255],
                [0, 0, 0, 255],
                [0, 0, 0, 255],
            ],
            [
                [0, 0, 0, 255],
                [0, 0, 0, 255],
                [255, 255, 255, 255],
                [255, 255, 255, 255],
            ],
            [
                [0, 0, 0, 255],
                [0, 0, 0, 255],
                [255, 255, 255, 255],
                [255, 255, 255, 255],
            ],
        ],
        np.uint8,
    )

height, width = data.shape[:2]
num_mips = max(width.bit_length(), height.bit_length())

v = Viewer(config=Config(window=False))

img = Image.from_data(
    v.device,
    data,
    ImageLayout.GENERAL,
    width,
    height,
    Format.R8G8B8A8_UNORM,
    ImageUsageFlags.TRANSFER_DST | ImageUsageFlags.SAMPLED | ImageUsageFlags.STORAGE | ImageUsageFlags.TRANSFER_SRC,
    AllocType.DEVICE,
    create_flags=ImageCreateFlags.MUTABLE_FORMAT,
    mip_levels=num_mips,
)

v.renderer.spd_pipeline.run_sync(v.renderer, img, ImageLayout.TRANSFER_SRC_OPTIMAL, MipGenerationFilter.AVERAGE)

mips = readback_image_mips(v.device, img, ImageLayout.SHADER_READ_ONLY_OPTIMAL)
for m in mips:
    print(m.shape)
    print(m)



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Optional

import numpy as np
from pyxpg import (
    AccessFlags,
    AllocType,
    Buffer,
    BufferUsageFlags,
    CommandBuffer,
    ComputePipeline,
    DescriptorSetBinding,
    DescriptorType,
    DeviceFeatures,
    MemoryBarrier,
    MemoryUsage,
    PipelineStageFlags,
    PushConstantsRange,
    Shader,
)

from . import renderer
from .utils.descriptors import create_descriptor_layout_pool_and_sets_ringbuffer
from .utils.gpu import div_round_up

RADIX_SORT_BITS = 8
RADIX_SORT_RADIX = 1 << RADIX_SORT_BITS
RADIX_SORT_PASSES = 4
RADIX_SORT_GLOBAL_HIST_PARTITION_SIZE = 32768


class SortDataType(Enum):
    FLOAT32 = auto()
    INT32 = auto()
    UINT32 = auto()


# TODO: tuning based on vendor ID with fallback to default
# and validation on pipeline creation.
#
# Found so far:
#
# Apple M4 Max:
# keys_per_thread = 7,
# threads_per_threadblock = 256,
# partition_size = 7 * 256,
# total_shared_memory = 4096,
#
# NVDIA RTX 3060 Mobile:
# keys_per_thread = 15
# threads_per_threadblock = 256
# partition_size = 3840
# total_shared_memory = 4096
@dataclass(frozen=True)
class SortTuningParameters:
    lock_wave32: bool = False
    keys_per_thread: int = 15
    threads_per_threadblock: int = 256
    partition_size: int = 3840
    total_shared_memory: int = 4096


@dataclass(frozen=True)
class SortOptions:
    key_type: SortDataType
    payload_type: Optional[SortDataType] = None
    max_count: int = 256 * 1024 * 1024  # Max number of elements that can be sorted
    descending: bool = False  # If True sort in descending order. Otherwise ascending.
    unsafe_has_forward_thread_progress_guarantee: bool = False
    indirect: bool = False  # If True sort will be dispatched indirectly (currently direct dispatch is not allowed if this is set, but could be extended to allow both)


@dataclass
class DeviceRadixSortPipeline:
    init: ComputePipeline
    upsweep: ComputePipeline
    scan: ComputePipeline
    downsweep: ComputePipeline


@dataclass
class OnesweepPipeline:
    init: ComputePipeline
    global_histogram: ComputePipeline
    scan: ComputePipeline
    onesweep: ComputePipeline


class GpuSortingPipeline:
    def __init__(self, r: "renderer.Renderer", options: SortOptions, tuning: Optional[SortTuningParameters] = None):
        self.options = options
        self.tuning = tuning or SortTuningParameters()

        self.constants_dtype = np.dtype(
            {
                "numKeys": (np.uint32, 0),
                "radixShift": (np.uint32, 4),
                "threadBlocks": (np.uint32, 8),
                "isPartial": (np.uint32, 12),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)

        # NOTE: could switch to a different descriptor set just for the init kernel to use less max slots
        descriptors = [
            DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 0 - Keys
            DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 1 - Alt-keys
            DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 2 - Payload
            DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 3 - Alt-payload
            DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 4 - Global hist
            DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 5 - Pass hist
            DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 6 - Index (sweep only)
        ]

        if options.indirect:
            descriptors.extend(
                [
                    DescriptorSetBinding(1, DescriptorType.UNIFORM_BUFFER),  #  7 - indirect constants
                    DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  #  8 - indirect constants write
                    DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  #  9 - indirect dispatch write
                    DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 10 - indirect num keys
                ]
            )

        # Allocate 2 sets per frame to ping-pong radix passes
        self.descriptor_set_layout, self.descriptor_pool, self.descriptor_sets = (
            create_descriptor_layout_pool_and_sets_ringbuffer(
                r.device,
                descriptors,
                2 * r.num_frames_in_flight,
            )
        )

        enable_16_bit = (r.device.features & DeviceFeatures.SHADER_INT16) != 0

        # Permutations are:
        # 3 key types
        # 4 payload types (3 dtypes + 1 if not sorting pairs)
        # 2 sorting directions
        # 2 modes with 16 bits or 32 bits storage
        # Total:
        #   48 permutations max
        #   12 permutations if never sorting pairs
        #    6 permutations if never sorting descending
        #    3 permutations if never using 16 bit storage
        defines = []
        if options.key_type == SortDataType.FLOAT32:
            defines.append(("KEY_FLOAT", ""))
        elif options.key_type == SortDataType.INT32:
            defines.append(("KEY_INT", ""))
        elif options.key_type == SortDataType.UINT32:
            defines.append(("KEY_UINT", ""))
        else:
            raise ValueError(f"Unhandled key data type: {options.key_type}")

        if options.payload_type is not None:
            defines.append(("SORT_PAIRS", ""))
            if options.payload_type == SortDataType.FLOAT32:
                defines.append(("PAYLOAD_FLOAT", ""))
            elif options.payload_type == SortDataType.INT32:
                defines.append(("PAYLOAD_INT", ""))
            elif options.payload_type == SortDataType.UINT32:
                defines.append(("PAYLOAD_UINT", ""))
            else:
                raise ValueError(f"Unhandled payload data type: {options.payload_type}")
        else:
            defines.append(("PAYLOAD_UINT", ""))

        if not options.descending:
            defines.append(("SHOULD_ASCEND", ""))

        if enable_16_bit:
            defines.append(("ENABLE_16_BIT", ""))

        defines.append(("MAX_DISPATCH_DIM", str(r.device.device_properties.limits.max_compute_work_group_count[0])))
        if options.indirect:
            defines.append(("INDIRECT_DISPATCH", ""))

        # Tuning
        defines.append((f"KEYS_PER_THREAD_{self.tuning.keys_per_thread}", ""))
        defines.append((f"PART_SIZE_{self.tuning.partition_size}", ""))
        defines.append((f"D_TOTAL_SMEM_{self.tuning.total_shared_memory}", ""))
        defines.append((f"D_DIM_{self.tuning.threads_per_threadblock}", ""))

        def create_pipeline(shader_name: str, entry: str) -> ComputePipeline:
            shader = r.compile_builtin_shader(
                Path("GPUSorting", shader_name),
                entry=entry,
                defines=defines,
                include_paths=[renderer.SHADERS_PATH.joinpath("GPUSorting")],
            )
            return ComputePipeline(
                r.device,
                Shader(r.device, shader.code),
                descriptor_set_layouts=[self.descriptor_set_layout],
                push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
            )

        if options.unsafe_has_forward_thread_progress_guarantee:
            self.onesweep_pipeline = OnesweepPipeline(
                init=create_pipeline("SweepCommon.slang", "InitSweep"),
                global_histogram=create_pipeline("SweepCommon.slang", "GlobalHistogram"),
                scan=create_pipeline("SweepCommon.slang", "Scan"),
                onesweep=create_pipeline("OneSweep.slang", "DigitBinningPass"),
            )
        else:
            self.device_radix_sort_pipeline = DeviceRadixSortPipeline(
                init=create_pipeline("DeviceRadixSort.slang", "InitDeviceRadixSort"),
                upsweep=create_pipeline("DeviceRadixSort.slang", "Upsweep"),
                scan=create_pipeline("DeviceRadixSort.slang", "Scan"),
                downsweep=create_pipeline("DeviceRadixSort.slang", "Downsweep"),
            )

        self.global_histogram_buf = Buffer(
            r.device,
            RADIX_SORT_RADIX * RADIX_SORT_PASSES * 4,
            BufferUsageFlags.STORAGE,
            AllocType.DEVICE,
            name="gpu-sorting-global-histogram",
        )

        pass_histogram_buf_size = div_round_up(options.max_count, self.tuning.partition_size) * RADIX_SORT_RADIX * 4
        if options.unsafe_has_forward_thread_progress_guarantee:
            pass_histogram_buf_size *= RADIX_SORT_PASSES

        self.pass_histogram_buf = Buffer(
            r.device,
            pass_histogram_buf_size,
            BufferUsageFlags.STORAGE,
            AllocType.DEVICE,
            name="gpu-sorting-pass-histogram",
        )
        if options.unsafe_has_forward_thread_progress_guarantee:
            self.index_buf = Buffer(
                r.device,
                RADIX_SORT_PASSES * 4,
                BufferUsageFlags.STORAGE,
                AllocType.DEVICE,
                name="gpu-sorting-index",
            )

        if options.indirect:
            self.indirect_constants_buf = Buffer(
                r.device,
                16,
                BufferUsageFlags.STORAGE | BufferUsageFlags.UNIFORM,
                AllocType.DEVICE,
                name="gpu-sorting-indirect-constants",
            )

            self.indirect_dispatch_buf = Buffer(
                r.device,
                48 if options.unsafe_has_forward_thread_progress_guarantee else 24,
                BufferUsageFlags.STORAGE | BufferUsageFlags.INDIRECT,
                AllocType.DEVICE,
                name="gpu-sorting-indirect-dispatch-buf",
            )

        for s in self.descriptor_sets:
            s.write_buffer(self.global_histogram_buf, DescriptorType.STORAGE_BUFFER, 4)
            s.write_buffer(self.pass_histogram_buf, DescriptorType.STORAGE_BUFFER, 5)
            if options.unsafe_has_forward_thread_progress_guarantee:
                s.write_buffer(self.index_buf, DescriptorType.STORAGE_BUFFER, 6)
            if options.indirect:
                s.write_buffer(
                    self.indirect_constants_buf, DescriptorType.UNIFORM_BUFFER, 7
                )  # Indirect constants for reading
                s.write_buffer(
                    self.indirect_constants_buf, DescriptorType.STORAGE_BUFFER, 8
                )  # Indirect constants for writing in init shader
                s.write_buffer(
                    self.indirect_dispatch_buf, DescriptorType.STORAGE_BUFFER, 9
                )  # Indirect dispatch counts for writing in init shader

    def upload(
        self,
        r: "renderer.Renderer",
        key: Buffer,
        key_alt: Buffer,
        payload: Buffer,
        payload_alt: Buffer,
        indirect_count: Optional[Buffer],
    ) -> None:
        set0 = self.descriptor_sets.get_current_and_advance()
        set1 = self.descriptor_sets.get_current_and_advance()

        set0.write_buffer(key, DescriptorType.STORAGE_BUFFER, 0)
        set0.write_buffer(key_alt, DescriptorType.STORAGE_BUFFER, 1)
        set0.write_buffer(payload, DescriptorType.STORAGE_BUFFER, 2)
        set0.write_buffer(payload_alt, DescriptorType.STORAGE_BUFFER, 3)

        if indirect_count is not None:
            set0.write_buffer(indirect_count, DescriptorType.STORAGE_BUFFER, 10)

        set1.write_buffer(key_alt, DescriptorType.STORAGE_BUFFER, 0)
        set1.write_buffer(key, DescriptorType.STORAGE_BUFFER, 1)
        set1.write_buffer(payload_alt, DescriptorType.STORAGE_BUFFER, 2)
        set1.write_buffer(payload, DescriptorType.STORAGE_BUFFER, 3)

        self.current_sets = [set0, set1]

    def _run(
        self,
        r: "renderer.Renderer",
        cmd: CommandBuffer,
        count: int,
        indirect_count_offset: Optional[int],
    ) -> None:
        # Count exceeds max
        if self.options.max_count < count:
            raise ValueError(
                f"count ({count}) must be smaller than max_count ({self.options.max_count}) specified on sorting pipeline creation"
            )

        set_index = 0
        sets = self.current_sets

        thread_blocks = div_round_up(count, self.tuning.partition_size)

        max_dispatch_size = r.device.device_properties.limits.max_compute_work_group_count[0]
        full_blocks = thread_blocks // max_dispatch_size
        partial_blocks = thread_blocks - full_blocks * max_dispatch_size

        self.constants["numKeys"] = count if indirect_count_offset is None else (indirect_count_offset >> 2)
        self.constants["threadBlocks"] = thread_blocks

        if self.options.unsafe_has_forward_thread_progress_guarantee:
            # Onesweep sort

            self.constants["radixShift"] = 0

            # Init
            cmd.bind_descriptor_sets(self.onesweep_pipeline.init, [sets[0]])
            cmd.bind_pipeline(self.onesweep_pipeline.init)
            cmd.push_constants(self.onesweep_pipeline.init, self.constants.tobytes())
            cmd.dispatch(256, 1, 1)

            dst_stage = PipelineStageFlags.COMPUTE_SHADER
            dst_access = AccessFlags.SHADER_READ | AccessFlags.SHADER_WRITE
            if indirect_count_offset is not None:
                dst_stage |= PipelineStageFlags.DRAW_INDIRECT
                dst_access |= AccessFlags.INDIRECT_COMMAND_READ
            cmd.memory_barrier_full(
                MemoryBarrier(
                    PipelineStageFlags.COMPUTE_SHADER,
                    AccessFlags.SHADER_READ | AccessFlags.SHADER_WRITE,
                    dst_stage,
                    dst_access,
                )
            )

            # Global hist
            cmd.bind_pipeline(self.onesweep_pipeline.global_histogram)

            if indirect_count_offset is not None:
                self.constants["isPartial"] = 0
                cmd.push_constants(self.onesweep_pipeline.global_histogram, self.constants.tobytes())
                cmd.dispatch_indirect(self.indirect_dispatch_buf, 0)

                self.constants["isPartial"] = 1
                cmd.push_constants(self.onesweep_pipeline.global_histogram, self.constants.tobytes())
                cmd.dispatch_indirect(self.indirect_dispatch_buf, 12)
            else:
                global_hist_thread_blocks = div_round_up(count, RADIX_SORT_GLOBAL_HIST_PARTITION_SIZE)
                global_hist_full_blocks = global_hist_thread_blocks // max_dispatch_size
                global_hist_partial_blocks = global_hist_thread_blocks - global_hist_full_blocks * max_dispatch_size
                self.constants["threadBlocks"] = global_hist_thread_blocks

                if full_blocks > 0:
                    self.constants["isPartial"] = 0
                    cmd.push_constants(self.onesweep_pipeline.global_histogram, self.constants.tobytes())
                    cmd.dispatch(max_dispatch_size, global_hist_full_blocks, 1)

                if partial_blocks > 0:
                    self.constants["isPartial"] = (full_blocks << 1) | 1
                    cmd.push_constants(self.onesweep_pipeline.global_histogram, self.constants.tobytes())
                    cmd.dispatch(global_hist_partial_blocks, 1, 1)

            cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

            # Scan
            cmd.bind_pipeline(self.onesweep_pipeline.scan)
            self.constants["threadBlocks"] = thread_blocks
            cmd.push_constants(self.onesweep_pipeline.scan, self.constants.tobytes())
            cmd.dispatch(RADIX_SORT_PASSES, 1, 1)
            cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

            # Sweep
            self.constants["isPartial"] = 0
            cmd.bind_pipeline(self.onesweep_pipeline.onesweep)
            for radix_shift in range(0, 32, 8):
                # Bind descriptor set (skip first because already bound)
                if radix_shift > 0:
                    cmd.bind_descriptor_sets(self.onesweep_pipeline.onesweep, [sets[set_index]])
                set_index = 1 - set_index

                self.constants["radixShift"] = radix_shift
                cmd.push_constants(self.onesweep_pipeline.onesweep, self.constants.tobytes())

                if indirect_count_offset is not None:
                    cmd.dispatch_indirect(self.indirect_dispatch_buf, 24)
                    cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)
                    cmd.dispatch_indirect(self.indirect_dispatch_buf, 36)
                else:
                    if full_blocks > 0:
                        cmd.dispatch(max_dispatch_size, full_blocks, 1)
                        # Need a memory barrier between passes because they are dependent
                        cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

                    if partial_blocks > 0:
                        cmd.dispatch(partial_blocks, 1, 1)

                # Skip barrier for last iteration
                if radix_shift < 24:
                    cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

        else:
            # Device radix sort

            # Init
            cmd.bind_descriptor_sets(self.device_radix_sort_pipeline.init, [sets[0]])
            cmd.push_constants(self.device_radix_sort_pipeline.init, self.constants.tobytes())
            cmd.bind_pipeline(self.device_radix_sort_pipeline.init)
            cmd.dispatch(1, 1, 1)

            dst_stage = PipelineStageFlags.COMPUTE_SHADER
            dst_access = AccessFlags.SHADER_READ | AccessFlags.SHADER_WRITE
            if indirect_count_offset is not None:
                dst_stage |= PipelineStageFlags.DRAW_INDIRECT
                dst_access |= AccessFlags.INDIRECT_COMMAND_READ
            cmd.memory_barrier_full(
                MemoryBarrier(
                    PipelineStageFlags.COMPUTE_SHADER,
                    AccessFlags.SHADER_READ | AccessFlags.SHADER_WRITE,
                    dst_stage,
                    dst_access,
                )
            )

            for radix_shift in range(0, 32, 8):
                # Bind descriptor set (skip first because already bound)
                if radix_shift > 0:
                    cmd.bind_descriptor_sets(self.device_radix_sort_pipeline.upsweep, [sets[set_index]])
                set_index = 1 - set_index

                # Upsweep
                self.constants["radixShift"] = radix_shift

                cmd.bind_pipeline(self.device_radix_sort_pipeline.upsweep)
                if indirect_count_offset is not None:
                    self.constants["isPartial"] = 0
                    cmd.push_constants(self.device_radix_sort_pipeline.upsweep, self.constants.tobytes())
                    cmd.dispatch_indirect(self.indirect_dispatch_buf, 0)

                    self.constants["isPartial"] = 1
                    cmd.push_constants(self.device_radix_sort_pipeline.upsweep, self.constants.tobytes())
                    cmd.dispatch_indirect(self.indirect_dispatch_buf, 12)
                else:
                    if full_blocks > 0:
                        self.constants["isPartial"] = 0
                        cmd.push_constants(self.device_radix_sort_pipeline.upsweep, self.constants.tobytes())
                        cmd.dispatch(max_dispatch_size, full_blocks, 1)

                    if partial_blocks > 0:
                        self.constants["isPartial"] = (full_blocks << 1) | 1
                        cmd.push_constants(self.device_radix_sort_pipeline.upsweep, self.constants.tobytes())
                        cmd.dispatch(partial_blocks, 1, 1)

                cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

                # Scan
                cmd.bind_pipeline(self.device_radix_sort_pipeline.scan)
                cmd.dispatch(256, 1, 1)

                cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

                # Downsweep
                cmd.bind_pipeline(self.device_radix_sort_pipeline.downsweep)
                if indirect_count_offset is not None:
                    self.constants["isPartial"] = 0
                    cmd.push_constants(self.device_radix_sort_pipeline.downsweep, self.constants.tobytes())
                    cmd.dispatch_indirect(self.indirect_dispatch_buf, 0)

                    self.constants["isPartial"] = 1
                    cmd.push_constants(self.device_radix_sort_pipeline.downsweep, self.constants.tobytes())
                    cmd.dispatch_indirect(self.indirect_dispatch_buf, 12)
                else:
                    if full_blocks > 0:
                        self.constants["isPartial"] = 0
                        cmd.push_constants(self.device_radix_sort_pipeline.downsweep, self.constants.tobytes())
                        cmd.dispatch(max_dispatch_size, full_blocks, 1)

                    if partial_blocks > 0:
                        self.constants["isPartial"] = (full_blocks << 1) | 1
                        cmd.push_constants(self.device_radix_sort_pipeline.downsweep, self.constants.tobytes())
                        cmd.dispatch(partial_blocks, 1, 1)

                # Skip barrier for last iteration
                if radix_shift < 24:
                    cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)

    def run(
        self,
        r: "renderer.Renderer",
        cmd: CommandBuffer,
        count: int,
    ) -> None:
        if self.options.indirect:
            raise RuntimeError("GpuSortingPipeline was created for indirect dispatch. Direct dispatch is not allowed.")

        # Nothing to do
        if count <= 1:
            return

        self._run(r, cmd, count, None)

    def run_indirect(
        self,
        r: "renderer.Renderer",
        cmd: CommandBuffer,
        count_offset: int,
    ) -> None:
        if not self.options.indirect:
            raise RuntimeError("GpuSortingPipeline was created for direct dispatch. Indirect dispatch is not allowed.")
        self._run(r, cmd, 0, count_offset)
